﻿using System;
using System.Text;

using Arinc.Applications.Domain.Graphics.Interfaces;

using ThinkGeo.Core;

namespace Arinc.Applications.Domain.Graphics.Pocos
{
    /// <summary>
    /// Base definition for a graphic background image
    /// </summary>
    public abstract class BaseGraphicBackgroundImage : BaseGraphicImage, IGraphicBackgroundImage
    {
        #region Members

        private string _backgroundColorName;
        private GeographyUnit _mapUnit;

        #endregion // Members

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="key">Unique key of the graphic image</param>
        /// <param name="name">The display name of the graphic image</param>
        /// <param name="filename">Graphic image filename</param>
        /// <param name="backgroundColorName">Color to display on the graphic outside the bounds of the image</param>
        /// <param name="mapUnit">The units to use when displaying the background image</param>
        /// <param name="minX">Minimum X coordinate</param>
        /// <param name="maxY">Maximum Y coordinate</param>
        /// <param name="maxX">Maximum X coordinate</param>
        /// <param name="minY">Minimum X coordinate</param>
        public BaseGraphicBackgroundImage(
                    string key,
                    string name,
                    string filename,
                    string backgroundColorName,
                    GeographyUnit mapUnit,
                    double minX,
                    double maxY,
                    double maxX,
                    double minY)
            : base(key, name, filename)
        {
            BackgroundColorName = backgroundColorName;
            MapUnit = mapUnit;
            MinX = minX;
            MaxY = maxY;
            MaxX = maxX;
            MinY = minY;
        }

        #endregion // Constructors

        #region Events
        #endregion // Events

        #region Properties

        #region Public Properties

        /// <summary>
        /// Gets/Sets the name of the background color
        /// </summary>
        public string BackgroundColorName
        {
            get
            {
                return _backgroundColorName;
            }
            set
            {
                _backgroundColorName = value;
                OnPropertyChanged("BackgroundColorName");
            }
        }

        /// <summary>
        /// Gets/Sets the view's map unit
        /// </summary>
        public GeographyUnit MapUnit
        {
            get
            {
                return _mapUnit;
            }
            set
            {
                _mapUnit = value;
                OnPropertyChanged("MapUnit");
            }
        }

        /// <summary>
        /// Gets/Sets the minimum X coordinate
        /// </summary>
        public double MinX { get; private set; }

        /// <summary>
        /// Gets/Sets the maximum Y coordinate
        /// </summary>
        public double MaxY { get; private set; }

        /// <summary>
        /// Gets/Sets the maximum X coordinate
        /// </summary>
        public double MaxX { get; private set; }

        /// <summary>
        /// Gets/Sets the minimum Y coordinate
        /// </summary>
        public double MinY { get; private set; }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Returns the image spatial location as the specified type "T"
        /// Throws an exception if the type is not supported
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public abstract T GetImageLocation<T>();

        /// <summary>
        /// Returns a string representation of the data
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();
            try
            {
                str.AppendFormat("Name: '{0}'  ", Name);
                str.AppendFormat("Key: '{0}'  ", Key);
                str.AppendFormat("Filename: '{0}'  ", Filename);
                str.AppendFormat("MapUnit: '{0}'  ", MapUnit);
                str.AppendFormat("MinX: '{0}'  ", MinX);
                str.AppendFormat("MaxY: '{0}'  ", MaxY);
                str.AppendFormat("MaxX: '{0}'  ", MaxX);
                str.AppendFormat("MinY: '{0}'  ", MinY);
            }
            catch (Exception ex)
            {
                str.AppendFormat("Exception - {0}", ex.Message);
            }

            return str.ToString();
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods
        #endregion // Private Methods

        #endregion // Methods
    }
}
