﻿using System;
using System.Text;
using System.ComponentModel;

using Arinc.Applications.Domain.Graphics.Interfaces;

namespace Arinc.Applications.Domain.Graphics.Pocos
{
    /// <summary>
    /// Base definition of a graphic image (e.g., background or symbol)
    /// </summary>
    public abstract class BaseGraphicImage : IGraphicImage, INotifyPropertyChanged
    {
        #region Members

        private string _key;
        private string _name;
        private string _filename;

        #endregion // Members

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="key">Unique key of the graphic image</param>
        /// <param name="name">The display name of the graphic image</param>
        /// <param name="filename">Graphic image filename</param>
        public BaseGraphicImage(string key, string name, string filename)
        {
            Key = key;
            Name = name;
            Filename = filename;
        }

        #endregion // Constructors

        #region Events

        public event PropertyChangedEventHandler PropertyChanged;
        
        #endregion // Events

        #region Properties

        #region Public Properties

        /// <summary>
        /// Gets the unique key of the graphic image (classname)
        /// </summary>
        public string Key
        {
            get
            {
                return _key;
            }
            set
            {
                _key = value;
                OnPropertyChanged("Key");
            }
        }

        /// <summary>
        /// Gets the display name of the graphic image
        /// </summary>
        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
                OnPropertyChanged("Name");
            }
        }

        /// <summary>
        /// Gets the graphic image filename
        /// </summary>
        public string Filename
        {
            get
            {
                return _filename;
            }
            set
            {
                _filename = value;
                OnPropertyChanged("Filename");
            }
        }

        /// <summary>
        /// Height of the image
        /// </summary>
        public abstract float Height { get; }

        /// <summary>
        /// Width of the image
        /// </summary>
        public abstract float Width { get; }

        #endregion // Public Properties

        #region Protected Properties

        protected void OnPropertyChanged(string name)
        {
            PropertyChangedEventHandler subscribers = PropertyChanged;
            if (subscribers != null)
            {
                subscribers(this, new PropertyChangedEventArgs(name));
            }
        }

        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Gets the image
        /// </summary>
        public abstract T GetImage<T>();

        /// <summary>
        /// Compares the key of this graphic image with another graphic image
        /// </summary>
        /// <param name="other"></param>
        /// <returns></returns>
        public bool Equals(IGraphicImage other)
        {
            if (other != null)
            {
                return (string.Compare(Key, other.Key, true) == 0);
            }

            return false;
        }

        public override bool Equals(object obj)
        {
            if (obj is IGraphicImage)
            {
                IGraphicImage other = obj as IGraphicImage;
                return string.Compare(Key, other.Key, true) == 0;
            }

            return base.Equals(obj);
        }

        public override int GetHashCode()
        {
            return Key.GetHashCode();
        }

        /// <summary>
        /// ToString() override
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();
            try
            {
                str.Append("BaseGraphicImage - ");
                str.AppendFormat("Name: '{0}'  ", Name);
                str.AppendFormat("Key: '{0}'  ", Key);
                str.AppendFormat("Filename: '{0}'  ", Filename);
            }
            catch (Exception ex)
            {
                str.AppendFormat("Exception - {0}", ex.Message);
            }

            return str.ToString();
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods
        #endregion // Private Methods

        #endregion // Methods
    }
}
