﻿using Arinc.Applications.Domain.Graphics.Interfaces;

using System;
using System.IO;
using System.Text;

namespace Arinc.Applications.Domain.Graphics.Pocos
{
    /// <summary>
    /// Base definition of a graphic image stored in the database (e.g., background or symbol)
    /// </summary>
    public abstract class BaseGraphicImageData
    {
        #region Members

        private string _filename;
        private byte[] _imageBytes;

        #endregion // Members

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public BaseGraphicImageData()
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="id">Id of the image</param>
        /// <param name="name">Name of the image</param>
        /// <param name="description">Description of the image</param>
        /// <param name="filename">Filename the image was loaded from</param>
        /// <param name="imageBytes">Image as an array of bytes</param>
        public BaseGraphicImageData(Guid id, string name, string description, string filename, byte[] imageBytes)
        {
            Id = id;
            Name = name;

            _filename = filename;
            _imageBytes = imageBytes;
        }

        #endregion // Constructors

        #region Events
        #endregion // Events

        #region Properties

        #region Public Properties

        public Guid Id
        {
            get;
            set;
        }

        public virtual string Name
        {
            get;
            set;
        }

        /// <summary>
        /// Gets the filename the image was loaded from
        /// </summary>
        public string Filename
        {
            get { return _filename; }
            set { _filename = value; }
        }

        /// <summary>
        /// Gets the image as an array of bytes
        /// </summary>
        public byte[] ImageBytes
        {
            get { return _imageBytes; }
            set { _imageBytes = value; }
        }

        /// <summary>
        /// Gets the unique key used to reference the image in the graphic image library
        /// </summary>
        public string Key
        {
            get
            {
                return Id.ToString();
            }
            private set
            {
                
            }
        }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Returns the image in the specified type
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public virtual T GetImage<T>()
        {
            if (typeof(T).Equals(typeof(Stream)))
            {
                object ms = new MemoryStream(ImageBytes);
                return (T)ms;
            }
            else if (typeof(T).Equals(typeof(byte[])))
            {
                object bytes = ImageBytes;
                return (T)bytes;
            }

            throw new ArgumentException(string.Format("Image type '{0}' is not supported - Currently only support 'MemoryStream'", typeof(T).Name));
        }


        /// <summary>
        /// Returns true if this image is the same as the specified image
        /// </summary>
        /// <param name="other"></param>
        /// <returns></returns>
        public bool Equals(IGraphicImage other)
        {
            if (other != null)
            {
                return (string.Compare(Key, other.Key, true) == 0);
            }

            return false;
        }

        /// <summary>
        /// Returns a string representation of the data
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();
            try
            {
                str.Append("BaseGraphicImageData - ");
                str.AppendFormat("Name: '{0}'  ", Name);
                str.AppendFormat("Id: '{0}'  ", Id);
                str.AppendFormat("Filename: '{0}'  ", Filename);
                str.AppendFormat("ImageBytes: '{0}'  ", ImageBytes != null ? ImageBytes.Length.ToString() : "0");
            }
            catch (Exception ex)
            {
                str.AppendFormat("Exception - {0}", ex.Message);
            }

            return str.ToString();
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods
        #endregion // Private Methods

        #endregion // Methods
    }
}
