﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using Arinc.Applications.Domain.Graphics.Interfaces;

using ThinkGeo.Core;

namespace Arinc.Applications.Domain.Graphics.Pocos
{
    /// <summary>
    /// Represents a raster image that can be displayed as a background on a graphic
    /// </summary>
    public class GraphicBackgroundRasterImage : BaseGraphicBackgroundImage, IGraphicBackgroundRasterImage
    {
        #region Members

        private byte[] _imageBytes;
        private float _imageHeight = -1;
        private float _imageWidth = -1;
        private string _imageLocation;
        private byte[] _imageLocationBytes;

        #endregion // Members

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="key">Unique key of the graphic image</param>
        /// <param name="name">The display name of the graphic image</param>
        /// <param name="filename">Graphic image filename</param>
        /// <param name="imageBytes">Background image as a array of bytes</param>
        /// <param name="backgroundColorName">Color to display on the graphic outside the bounds of the image</param>
        /// <param name="mapUnit">The units to use when displaying the background image</param>
        /// <param name="minX">Minimum X coordinate</param>
        /// <param name="maxY">Maximum Y coordinate</param>
        /// <param name="maxX">Maximum X coordinate</param>
        /// <param name="minY">Minimum X coordinate</param>
        public GraphicBackgroundRasterImage(
                    string key,
                    string name,
                    string filename,
                    byte [] imageBytes,
                    string backgroundColorName,
                    GeographyUnit mapUnit,
                    double minX,
                    double maxY,
                    double maxX,
                    double minY)
            : base(key, name, filename, backgroundColorName, mapUnit, minX, maxY, maxX, minY)
        {
            _imageBytes = imageBytes != null ? imageBytes : CreateDefaultImage(Math.Abs(maxX - minX), Math.Abs(maxY - minY));
            CalculateImageLocation(MinX, MaxY, MaxX, MinY);
        }

        #endregion // Constructors

        #region Events
        #endregion // Events

        #region Properties

        #region Public Properties

        /// <summary>
        /// Gets the background image as a array of bytes
        /// </summary>
        public byte[] ImageBytes
        {
            get
            {
                return _imageBytes;
            }
        }

        /// <summary>
        /// Gets the world location "file" that defines the image spatially
        /// </summary>
        public byte[] ImageLocationBytes
        {
            get
            {
                return _imageLocationBytes;
            }
        }

        /// <summary>
        /// Height of the image
        /// </summary>
        public override float Height
        {
            get
            {
                if (_imageHeight <= 0)
                {
                    CalculateImageWidthAndHeight();
                }

                return _imageHeight;
            }
        }

        /// <summary>
        /// Width of the image
        /// </summary>
        public override float Width
        {
            get
            {
                if (_imageWidth <= 0)
                {
                    CalculateImageWidthAndHeight();
                }

                return _imageWidth;
            }
        }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Returns the image in the specified type
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public override T GetImage<T>()
        {
            if (typeof(T).Equals(typeof(Stream)))
            {
                object ms = new MemoryStream(ImageBytes);
                return (T)ms;
            }
            else if (typeof(T).Equals(typeof(byte[])))
            {
                object bytes = ImageBytes;
                return (T)bytes;
            }

            throw new ArgumentException(string.Format("Image type '{0}' is not supported - Currently only support 'Stream' (as a MemoryStream) and 'byte[]'", typeof(T).Name));
        }

        /// <summary>
        /// Returns the image spatial location as the specified type "T"
        /// Throws an exception if the type is not supported
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        public override T GetImageLocation<T>()
        {
            if (typeof(T).Equals(typeof(Stream)))
            {
                object ms = new MemoryStream(ImageLocationBytes);
                return (T)ms;
            }
            else if (typeof(T).Equals(typeof(byte[])))
            {
                object bytes = ImageLocationBytes;
                return (T)bytes;
            }
            else if (typeof(T).Equals(typeof(string)))
            {
                object location = _imageLocation;
                return (T)location;
            }

            throw new ArgumentException(string.Format("Image type '{0}' is not supported - Currently only support 'Stream' (as a MemoryStream), 'byte[]' and 'string'", typeof(T).Name));
        }

        /// <summary>
        /// Returns a string representation of the data
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();
            try
            {
                str.Append("GraphicBackgroundRasterImage - ");
                str.AppendFormat("{0}  ", base.ToString());
                str.AppendFormat("ImageBytes Count: '{0}'  ", _imageBytes?.Length);
                str.AppendFormat("Height: '{0}'  ", Height);
                str.AppendFormat("Width: '{0}'  ", Width);
                str.AppendFormat("ImageLocation: '{0}'  ", _imageLocation?.Replace("\r\n", "|"));
            }
            catch (Exception ex)
            {
                str.AppendFormat("Exception - {0}", ex.Message);
            }

            return str.ToString();
        }

        /// <summary>
        /// Converts a GraphicViewBackgroundData to an IGraphicBackgroundRasterImage
        /// </summary>
        /// <param name="backgroundData"></param>
        /// <returns></returns>
        public static IGraphicBackgroundRasterImage CreateFrom(GraphicViewBackgroundData backgroundData)
        {
            IGraphicBackgroundRasterImage backgroundImage = null;
            if (backgroundData != null)
            {
                backgroundImage =
                    new GraphicBackgroundRasterImage(
                            backgroundData.Id.ToString(),
                            backgroundData.Name,
                            backgroundData.Filename,
                            backgroundData.ImageBytes,
                            backgroundData.BackgroundColorName,
                            backgroundData.MapUnit,
                            backgroundData.MinX,
                            backgroundData.MaxY,
                            backgroundData.MaxX,
                            backgroundData.MinY);
            }

            return backgroundImage;
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods

        /// <summary>
        /// Create a transparent bitmap using the specified width and height in pixels
        /// </summary>
        /// <returns></returns>
        private byte[] CreateDefaultImage(double width, double height)
        {
            byte[] imageBytes = new byte[0];

            using (Bitmap image = new Bitmap((int)width, (int)height))
            {
                // Convert the image to bytes
                using (Stream imageStream = new MemoryStream())
                {
                    image.MakeTransparent();
                    image.Save(imageStream, ImageFormat.Png);
                    imageStream.Position = 0;
                    imageBytes = new byte[imageStream.Length];
                    imageStream.Read(imageBytes, 0, imageBytes.Length);
                }
            }

            return imageBytes;
        }

        /// <summary>
        /// Calculates the spatial location of the image based on its size
        /// </summary> 
        /// <param name="minX">Upper left X</param>
        /// <param name="maxY">Upper left Y</param>
        /// <param name="maxX">Lower right X</param>
        /// <param name="minY">Lower right Y</param>
        /// <returns>The spatial/world location of the image</returns>
        private void CalculateImageLocation(double minX, double maxY, double maxX, double minY)
        {
            double pixelSize_X = (maxX - minX) / Width;
            double rotation_Y = 0.0;
            double rotation_X = 0.0;
            double pixelSize_Y = -((maxY - minY) / Height);
            double upperLeftPixel_X = minX;
            double upperLeftPixel_Y = maxY;

            _imageLocation = String.Format("{0}\r\n{1}\r\n{2}\r\n{3}\r\n{4}\r\n{5}", pixelSize_X, rotation_Y, rotation_X, pixelSize_Y, upperLeftPixel_X, upperLeftPixel_Y);
            _imageLocationBytes = Encoding.UTF8.GetBytes(_imageLocation);
        }

        /// <summary>
        /// Calculates the height and width of the image
        /// </summary>
        private void CalculateImageWidthAndHeight()
        {
            GraphicImageHelper.CalculateRasterImageWidthAndHeight(ImageBytes, out _imageWidth, out _imageHeight);
        }

        #endregion // Private Methods

        #endregion // Methods
    }
}
