﻿using System;
using System.IO;
using System.Text;

using ThinkGeo.Core;

using Arinc.Applications.Domain.Graphics.Pocos;
using Arinc.Applications.Domain.Graphics.Interfaces;

namespace Arinc.UI.Vendor.ThinkGeo.Poco
{
    /// <summary>
    /// Custom class that keeps the background image and image location streams
    /// </summary>
    public class GraphicRasterLayer : NativeImageRasterLayer
    {
        #region Members

        private IGraphicBackgroundImage _backgroundImage;
        private GraphicViewBackgroundData _backgroundData;
        private static IGraphicBackgroundRasterImage _defaultBackgroundImage;
        private string _backgroundColorName = "Black";

        #endregion // Members

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        private GraphicRasterLayer()
        {
        }

        public GraphicRasterLayer(GraphicViewBackgroundData backgroundData, IGraphicBackgroundImage backgroundImage)
        {
            _backgroundData = backgroundData;
            _backgroundImage = backgroundImage;

            InitializeImage();
        }

        #endregion // Constructors

        #region Events
        #endregion // Events

        #region Properties

        #region Public Properties

        /// <summary>
        /// Gets/sets the map unit for this custom layer
        /// </summary>
        public GeographyUnit LayerMapUnit { get; private set; }

        /// <summary>
        /// Gets the layer's background color
        /// </summary>
        public string BackgroundColorName
        {
            get
            {
                return _backgroundData != null ? _backgroundData.BackgroundColorName : _backgroundColorName;
            }
        }

        /// <summary>
        /// Extent of the layer based on the size of the image
        /// </summary>
        public RectangleShape Extent { get; private set; }

        /// <summary>
        /// Gets the bytes that make up the image
        /// </summary>
        public byte[] ImageBytes
        {
            get
            {
                return _backgroundImage.GetImage<byte[]>();
            }
        }

        /// <summary>
        /// Gets the stream of the image
        /// </summary>
        public Stream ImageStream
        {
            get
            {
                return new MemoryStream(ImageBytes);
            }
        }

        /// <summary>
        /// Gets the bytes for the location of the image on the graphic
        /// </summary>
        public byte[] ImageLocationBytes
        {
            get
            {
                return _backgroundImage.GetImageLocation<byte[]>();
            }
        }

        /// <summary>
        /// Gets the memory stream containing the "world file" of the image
        /// </summary>
        public MemoryStream ImageLocationStream
        {
            get
            {
                return new MemoryStream(ImageLocationBytes);
            }
        }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties

        /// <summary>
        /// Gets the default background image to use for graphics that 
        /// do not have an explicit background image to display
        /// </summary>
        private IGraphicBackgroundImage DefaultBackgroundImage
        {
            get
            {
                if (_defaultBackgroundImage == null)
                {
                    _defaultBackgroundImage =
                        new GraphicBackgroundRasterImage(
                                Guid.NewGuid().ToString(),
                                "$DefaultBackgroundImage$",
                                string.Empty,
                                null,
                                _backgroundColorName,
                                GeographyUnit.Meter,
                                0.0,
                                1.0,
                                1.0,
                                0.0);
                }

                return _defaultBackgroundImage;
            }
        }

        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Returns a string representation of the data
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();

            try
            {
                str.Append("GraphicRasterLayer - ");
                str.AppendFormat("Name: '{0}'  ", Name);
                str.AppendFormat("ImagePathFilename: '{0}'  ", ImagePathFilename);
                str.AppendFormat("LayerMapUnit: '{0}'  ", LayerMapUnit);
                str.AppendFormat("BackgroundColorName: '{0}'  ", BackgroundColorName);
                str.AppendFormat("Extent: '{0}'  ", Extent.ToString());
                str.AppendFormat("ImageBytes Count: '{0}'  ", ImageBytes?.Length);
                str.AppendFormat("ImageLocation - {0}", _backgroundImage?.GetImageLocation<string>().Replace("\r\n", "|"));
            }
            catch (Exception ex)
            {
                str.AppendFormat("Exception - {0}", ex.ToString());
            }

            return str.ToString();
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods

        /// <summary>
        /// Initalizes the image for display
        /// </summary>
        private void InitializeImage()
        {
            Extent = new RectangleShape(_backgroundData.MinX, _backgroundData.MaxY, _backgroundData.MaxX, _backgroundData.MinY);
            LayerMapUnit = _backgroundData.MapUnit;
            ImagePathFilename = Name = _backgroundImage.Filename;

            // Subscribe to stream loading so we can pass ThinkGeo a stream to our image and spatial location
            StreamLoading += new EventHandler<StreamLoadingEventArgs>(RasterLayer_StreamLoading);
        }

        /// <summary>
        /// Event handler for streaming an image
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RasterLayer_StreamLoading(object sender, StreamLoadingEventArgs e)
        {
            if (e.StreamType == "Image File")
            {
                e.AlternateStream = ImageStream;
            }
            else if (e.StreamType == "World File")
            {
                e.AlternateStream = ImageLocationStream;
            }
            else
            {
                throw new ArgumentException( $"Unknown image stream type '{e.StreamType}'");
            }
        }

        #endregion // Private Methods

        #endregion // Methods
    }
}
