﻿using System;
using System.Text;

using ThinkGeo.Core;

namespace Arinc.Applications.Domain.Graphics.Pocos
{
    /// <summary>
    /// Defines a graphic background image saved in the database
    /// </summary>
    public class GraphicViewBackgroundData : BaseGraphicImageData
    {
        #region Members

        private string _backgroundColorName;

        private GeographyUnit _mapUnit;

        private double _minX;

        private double _maxY;

        private double _maxX;

        private double _minY;

        private bool _isDefaultBackgroundImage;

        #endregion // Members

        #region Constructors

        public GraphicViewBackgroundData()
            : base()
        {
            BackgroundColorName = "Black";
            MapUnit = GeographyUnit.Meter;
            MinX = 0.0;
            MaxY = 1000.0;
            MaxX = 1000.0;
            MinY = 0.0;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="id">Unique id of the background image</param>
        /// <param name="name">Name of the background image</param>
        /// <param name="description">Description of the background image</param>
        /// <param name="filename">Name of the file from which the background image was loaded</param>
        /// <param name="imageBytes">Background image as a array of bytes</param>
        /// <param name="backgroundColorName">Color to display on the graphic outside the bounds of the image</param>
        /// <param name="mapUnit">The units to use when displaying the background image</param>
        /// <param name="minX">Minimum X coordinate</param>
        /// <param name="maxY">Maximum Y coordinate</param>
        /// <param name="maxX">Maximum X coordinate</param>
        /// <param name="minY">Minimum X coordinate</param>
        public GraphicViewBackgroundData(
                        Guid id,
                        string name,
                        string description,
                        string filename,
                        byte[] imageBytes,
                        string backgroundColorName,
                        GeographyUnit mapUnit,
                        double minX,
                        double maxY,
                        double maxX,
                        double minY)
            : base(id, name, description, filename, imageBytes)
        {
            _backgroundColorName = backgroundColorName;
            _mapUnit = mapUnit;
            _minX = minX;
            _maxY = maxY;
            _maxX = maxX;
            _minY = minY;
        }

        #endregion // Constructors

        #region Events
        #endregion // Events

        #region Properties

        #region Public Properties

        /// <summary>
        /// Gets/Sets the font name 
        /// </summary>
        public override string Name
        {
            get { return base.Name; }

            set { base.Name = value; }
        }

        /// <summary>
        /// Gets/Sets the color to display on the graphic outside the bounds of the image
        /// </summary>
        public string BackgroundColorName
        {
            get { return _backgroundColorName; }
            set { _backgroundColorName = value; }
        }

        /// <summary>
        /// Gets/Sets the units to use when displaying the background image
        /// </summary>
        public GeographyUnit MapUnit
        {
            get { return _mapUnit; }
            set { _mapUnit = value; }
        }

        /// <summary>
        /// Gets/Sets Minimum X coordinate
        /// </summary>
        public double MinX
        {
            get { return _minX; }
            set { _minX = value; }
        }

        /// <summary>
        /// Gets/Sets Maximum Y coordinate
        /// </summary>
        public double MaxY
        {
            get { return _maxY; }
            set { _maxY = value; }
        }

        /// <summary>
        /// Gets/Sets Maximum X coordinate
        /// </summary>
        public double MaxX
        {
            get { return _maxX; }
            set { _maxX = value; }
        }

        /// <summary>
        /// Gets/Sets Minimum X coordinate
        /// </summary>
        public double MinY
        {
            get { return _minY; }
            set { _minY = value; }
        }

        /// <summary>
        /// Gets/Sets if this is the default background image initially assigned to all views
        /// </summary>
        public bool IsDefaultBackgroundImage
        {
            get
            {
                if (ImageBytes == null || _isDefaultBackgroundImage)
                {
                    return true;
                }

                return false;
            }
            set
            {
                _isDefaultBackgroundImage = value;
            }
        }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Returns true if this background image is the same as the specified background image 
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public override bool Equals(object obj)
        {
            if (obj is GraphicViewBackgroundData)
            {
                GraphicViewBackgroundData other = obj as GraphicViewBackgroundData;
                return string.Compare(Key, other.Key, true) == 0;
            }

            return base.Equals(obj);
        }

        /// <summary>
        /// Returns the background image hash code
        /// </summary>
        /// <returns></returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// Returns a string representation of the data
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder();

            try
            {
                str.Append("GraphicViewBackgroundData - ");
                str.AppendFormat("Name: '{0}'  ", Name);
                str.AppendFormat("Id: '{0}'  ", Id);
                str.AppendFormat("Filename: '{0}'  ", Filename);
                str.AppendFormat("ImageBytes: '{0}'  ", ImageBytes != null ? ImageBytes.Length.ToString() : "0");
                str.AppendFormat("BackgroundColorName: '{0}'  ", BackgroundColorName);
                str.AppendFormat("MapUnit: '{0}'  ", MapUnit);
                str.AppendFormat("Width: '{0}'  ", Math.Abs(MaxX - MinX));
                str.AppendFormat("Height: '{0}'  ", Math.Abs(MaxY - MinY));
                str.AppendFormat("MinX: '{0}'  ", MinX);
                str.AppendFormat("MaxY: '{0}'  ", MaxY);
                str.AppendFormat("MaxX: '{0}'  ", MaxX);
                str.AppendFormat("MinY: '{0}'  ", MinY);
                str.AppendFormat("{0}", base.ToString());
            }
            catch (Exception ex)
            {
                str.AppendFormat("Exception - {0}", ex.ToString());
            }

            return str.ToString();
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods
        #endregion // Private Methods

        #endregion // Methods
    }
}
