﻿using System;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;

namespace ThinkGeoPerformanceApp
{

    public delegate void FlashTimerTickDelegate(object sender, FlashTimerTickEventPayload e);

    public class FlashTimerTickEventPayload
    {
        public double Opacity { get; private set; }
        public FlashCommandEnum FlashCommand { get; private set; }

        public FlashTimerTickEventPayload(FlashCommandEnum flashCommand, double opacity)
        {
            FlashCommand = flashCommand;
            Opacity = opacity;
        }
    }

    public enum FlashCommandEnum
    {
        OFF = 0,
        ON = 1
    }

    /// <summary>
    /// Service class that fires an event at a specified interval read from a config file
    /// </summary>
    public class FlashTimer
    {
        #region Members

        private int _intervalInMilliseconds;
        private double _opacity;
        private FlashCommandEnum _lastFlashCommandSent;
        private Timer _flashTimer;
        private const int INTERVAL_MAX = 10000;
        private const int INTERVAL_MIN = 50;
        private const double OPACITY_MIN = 0.0;
        private const double OPACITY_MAX = 1.0;
        private int[] _timeSlices = new int[2];
        private Object timerLock = new Object();

        TaskFactory taskFactory;

        #endregion // Members

        #region Constructors

        /// <summary>
        /// Importing constructor
        /// </summary>
        public FlashTimer()
        {
            taskFactory = new TaskFactory(TaskScheduler.FromCurrentSynchronizationContext());
         
            Initialize();
        }

        #endregion // Constructors

        #region Events

        public event FlashTimerTickDelegate FlashTimerTick;

        #endregion // Events

        #region Properties

        #region Public Properties

        /// <summary>
        /// Get/set the interval between published events
        /// </summary>
        public int IntervalInMilliseconds
        {
            get
            {
                return _intervalInMilliseconds;
            }
            private set
            {
                _intervalInMilliseconds = value;

                //Bounds checking for the Interval
                _intervalInMilliseconds = Convert.ToInt32(Math.Floor((double)(_intervalInMilliseconds / INTERVAL_MIN)) * INTERVAL_MIN);
                _intervalInMilliseconds = _intervalInMilliseconds > INTERVAL_MAX ? INTERVAL_MAX : _intervalInMilliseconds;
                _intervalInMilliseconds = _intervalInMilliseconds < INTERVAL_MIN ? INTERVAL_MIN : _intervalInMilliseconds;

                _timeSlices[0] = _intervalInMilliseconds;
                _timeSlices[1] = _intervalInMilliseconds * 2;
            }
        }

        /// <summary>
        /// Get/set the opacity
        /// </summary>
        public double Opacity
        {
            get
            {
                return _opacity;
            }
            private set
            {
                _opacity = value;

                //Bounds checking for the Opacity
                _opacity = _opacity > OPACITY_MAX ? OPACITY_MAX : _opacity;
                _opacity = _opacity < OPACITY_MIN ? OPACITY_MIN : _opacity;
            }
        }

        /// <summary>
        /// Get/set the flashing command
        /// </summary>
        public FlashCommandEnum FlashCommand
        {
            get;
            private set;
        }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods

        /// <summary>
        /// Start the flash timer
        /// </summary>
        public void Start()
        {
            Stop();

            _flashTimer = new Timer(OnTimedEvent, null, IntervalInMilliseconds, Timeout.Infinite);
        }

        public void Stop()
        {
            if (_flashTimer != null)
            {
                _flashTimer.Dispose();
                _flashTimer = null;
            }
        }

        /// <summary>
        /// Event handler that is called at the frequency of the interval
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void OnTimedEvent(object state)
        {
            lock (timerLock)
            {
                Stopwatch watch = new Stopwatch();
                watch.Start();

                long utcNowMilliseconds = DateTime.UtcNow.Ticks/TimeSpan.TicksPerMillisecond;

                long normalizedTimeValueInMilliseconds = utcNowMilliseconds % _timeSlices[1];
                int currentTimeSlice = normalizedTimeValueInMilliseconds >= _timeSlices[0] ? 1 : 0;
                FlashCommand = currentTimeSlice == 0 ? FlashCommandEnum.OFF : FlashCommandEnum.ON;

                //Publish the event for flashing
                if (_lastFlashCommandSent != FlashCommand)
                {
                    _lastFlashCommandSent = FlashCommand;

                    taskFactory.StartNew(() =>
                    {
                        FlashTimerTick?.Invoke(this, new FlashTimerTickEventPayload(FlashCommand, Opacity));
                    });
                }

                long nextInterval = _timeSlices[currentTimeSlice] - (normalizedTimeValueInMilliseconds + watch.ElapsedMilliseconds);
                try
                {
                    _flashTimer.Change(nextInterval, Timeout.Infinite);
                }
                catch (Exception exc)
                {
                    Console.WriteLine("[FlashTimer.OnTimedEvent] Exception caught - '{0}'", exc.ToString());
                    _flashTimer.Change(IntervalInMilliseconds, Timeout.Infinite);
                }
            }
        }

        /// <summary>
        /// Event handler for initializing TimerService
        /// after importing all modules.
        /// </summary>
        public void OnImportsSatisfied()
        {

        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods

        /// <summary>
        /// Initialize the flash timer
        /// </summary>
        private void Initialize()
        {
            FlashCommand = FlashCommandEnum.OFF;
            IntervalInMilliseconds = 750;
            Opacity = 0.5;
        }
        #endregion // Private Methods

        #endregion // Methods
    }
}
