﻿using Arinc.UI.Common.Utility;

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Threading;

namespace ThinkGeoPerformanceApp
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        private TabItem _selectedTab;
        private int _rowCount = 1;
        private int _colCount = 1;

        private MemoryOptimizer _memoryOptimizer;
        bool _isMemoryOptimizationEnabled = false;
        private int _memoryOptimizationIntervalInSeconds = 10;
        private int _memoryOptimizationThresholdInMB = 1000;

        public event PropertyChangedEventHandler PropertyChanged;

        public List<BackgroundImage> BackgroundImageList { get; private set; } = new List<BackgroundImage>();

        public BackgroundImage SelectedBackgroundImage { get; set; }

        public static GraphicPointFeatureSource GlobalFeatureSource = new GraphicPointFeatureSource();

        public MainWindow()
        {
            LoadBackgroundImages();
            SelectedBackgroundImage = BackgroundImageList[0];

            this.DataContext = this;
            InitializeComponent();

            _memoryOptimizer = new MemoryOptimizer(MemoryOptimizationIntervalInSeconds, MemoryOptimizationThresholdInMB, IsMemoryOptimizationEnabled);
            GlobalFeatureSource.Open();
        }

        public int RowCount
        {
            get { return _rowCount; }
            set
            {
                _rowCount = value;
                OnPropertyChanged(nameof(RowCount));
            }
        }

        public int ColCount
        {
            get { return _colCount; }
            set
            {
                _colCount = value;
                OnPropertyChanged(nameof(ColCount));
            }
        }

        public bool IsMemoryOptimizationEnabled
        {
            get
            {
                return _isMemoryOptimizationEnabled;
            }
            set
            {
                _isMemoryOptimizationEnabled = value;
                _memoryOptimizer.IsEnabled = value;
            }
        }

        public int MemoryOptimizationIntervalInSeconds
        {
            get
            {
                return _memoryOptimizationIntervalInSeconds;
            }
            set
            {
                _memoryOptimizationIntervalInSeconds = value;
                _memoryOptimizer.OptimizationIntervalInSeconds = value;
            }
        }

        public int MemoryOptimizationThresholdInMB
        {
            get
            {
                return _memoryOptimizationThresholdInMB;
            }
            set
            {
                _memoryOptimizationThresholdInMB = value;
                _memoryOptimizer.OptimizationThresholdInMB = value;
            }
        }

        public TabItem SelectedTab
        {
            get
            {
                return _selectedTab;
            }
            set
            {
                MapControlView tab;

                if (_selectedTab != null)
                {
                    tab = _selectedTab.Content as MapControlView;
                    if (tab != null)
                    {
                        tab.IsSelected(false);
                    }
                }

                _selectedTab = value;
                OnPropertyChanged(nameof(SelectedTab));

                if (_selectedTab != null)
                {
                    tab = _selectedTab.Content as MapControlView;
                    if (tab != null)
                    {
                        tab.IsSelected(true);
                    }
                }
            }
        }

        private void OnPropertyChanged(string name)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(name));
        }

        private void AddTab_Click(object sender, RoutedEventArgs e)
        {
            int mapId = MapTabControl.Items.Count + 1;
            TabItem tabItem = new TabItem() { Header = $"Tab {mapId}", Content = new MapControlView(mapId, RowCount, ColCount, SelectedBackgroundImage) };
            MapTabControl.Items.Add(tabItem);

            this.Dispatcher.BeginInvoke(DispatcherPriority.Normal, new Action(() => MapTabControl.SelectedItem = tabItem));
        }

        private void LoadBackgroundImages()
        {
            //string filename;

            BackgroundImage backgroundImage = new BackgroundImage() { Name = "None", Filename = "None", ImageBytes = null };
            BackgroundImageList.Add(backgroundImage);

            //filename = "App_Data\\SiteOverviewSmall.jpg";
            //backgroundImage = new BackgroundImage() { Name = "Small (100KB)", Filename = filename, ImageBytes = File.ReadAllBytes(filename) };
            //BackgroundImageList.Add(backgroundImage);

            //filename = "App_Data\\SiteOverviewMedium.png";
            //backgroundImage = new BackgroundImage() { Name = "Medium (3MB)", Filename = filename, ImageBytes = File.ReadAllBytes(filename) };
            //BackgroundImageList.Add(backgroundImage);

            //filename = "App_Data\\SiteOverviewLarge.jpg";
            //backgroundImage = new BackgroundImage() { Name = "Large (13MB)", Filename = filename, ImageBytes = File.ReadAllBytes(filename) };
            //BackgroundImageList.Add(backgroundImage);

            //filename = "App_Data\\SiteOverviewHuge.png";
            //backgroundImage = new BackgroundImage() { Name = "Huge (29MB)", Filename = filename, ImageBytes = File.ReadAllBytes(filename) };
            //BackgroundImageList.Add(backgroundImage);
        }
    }

    public class BackgroundImage
    {
        public string Name { get; set; }
        public string Filename { get; set; }
        public byte[] ImageBytes { get; set; }
    }
}
