﻿using Arinc.Applications.Domain.Graphics.Interfaces;
using Arinc.Applications.Domain.Graphics.Pocos;
using Arinc.UI.Vendor.ThinkGeo.Poco;

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Timers;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

using ThinkGeo.Core;

namespace ThinkGeoPerformanceApp
{
    /// <summary>
    /// Interaction logic for MapControlView.xaml
    /// </summary>
    public partial class MapControlView : UserControl, INotifyPropertyChanged
    {
        private int _mapId;
        private int _numFeaturesToUpdate;
        private int _featureChangeIntervalMS;
        private BackgroundImage _backgroundImage;

        private double _distanceBetweenPointsX;
        private double _distanceBetweenPointsY;

        private double _featureWidth;
        private double _featureHeight;
        private double _zoomLevel20Scale;
        private double _zoomIncrement;

        public static bool _logDebug = false;
        private System.Timers.Timer _featureChangeTimer;
        private FlashTimer _flashTimer;
        private string _initialState = "0";
        private Random _random;
        private bool _randomize = true;
        private object _refreshActiveLock = new object();

        private int _totalPoints;
        private int _rowNumber = 0;
        private int _rowCount = 1;
        private int _colCount = 1;
        private int _colNumber = 0;
        private object _lock = new object();
        private GraphicViewOverlay _graphicViewOverlay;
        private string _featureOverlayName = "FeatureOverlay";
        private GraphicPointFeatureLayer _shapeLayer;
        private GraphicPointFeatureLayer _symbolLayer;
        private ObservableCollection<GraphicPointFeatureLayer> _featureLayers = new ObservableCollection<GraphicPointFeatureLayer>();
        private long totalRefreshMS = 0;
        private long totalRefreshCount = 0;
        private long totalUpdateMS = 0;
        private long totalUpdateCount = 0;
        private long _updateTime;
        private long _avgUpdateTime;
        private int _queueDepth;
        private bool _flash = false;
        private ThreadedWorkQueue<WorkItem> _workQueue;
        private List<MyBaseFeature> _flashFeatures;
        private DateTime _lastFlashTime;
        private bool _isFirstLoad = true;
        private bool _isSelected = false;
        private List<MyLineFeature> _lineFeatures = new List<MyLineFeature>();

        public const string COLUMN_ID = "Id";
        public const string COLUMN_STATE = "State";

        private enum FeatureType
        {
            PNGSymbol,
            PentagonShape,
            RectangleShape
        }

        public event PropertyChangedEventHandler PropertyChanged;

        public MapControlView()
        {
            this.DataContext = this;
            InitializeComponent();
        }

        public MapControlView(int mapId, int rowCount, int colCount, BackgroundImage backgroundImage)
            : this()
        {
            MapId = mapId;
            _rowCount = rowCount;
            _colCount = colCount;
            _backgroundImage = backgroundImage;
        }

        public static readonly DependencyProperty MapIdProperty =
            DependencyProperty.Register(
                                    "MapId",
                                    typeof(int),
                                    typeof(MapControlView),
                                    null);
        public int MapId
        {
            get { return (int)GetValue(MapIdProperty); }
            set { SetValue(MapIdProperty, value); }
        }

        public int UpdateRate
        {
            get
            {
                return _featureChangeIntervalMS;
            }
            set
            {
                _featureChangeIntervalMS = value;
                OnPropertyChanged("UpdateRate");
            }
        }

        public int NumFeaturesToUpdate
        {
            get
            {
                return _numFeaturesToUpdate;
            }
            set
            {
                _numFeaturesToUpdate = value;
                OnPropertyChanged("NumFeaturesToUpdate");
                InitializeFlashFeatures();
            }
        }

        public bool Randomize
        {
            get
            {
                return _randomize;
            }
            set
            {
                _randomize = value;
                OnPropertyChanged("Randomize");
            }
        }

        public int TotalPoints
        {
            get
            {
                return _totalPoints;
            }
            set
            {
                _totalPoints = value;
                OnPropertyChanged("TotalPoints");
            }
        }

        public ObservableCollection<GraphicPointFeatureLayer> MyFeatureLayers
        {
            get
            {
                return _featureLayers;
            }
        }

        public bool LogDebug
        {
            get
            {
                return _logDebug;
            }
            set
            {
                _logDebug = value;
                OnPropertyChanged("LogDebug");
            }
        }

        public bool Flash
        {
            get
            {
                return _flash;
            }
            set
            {
                _flash = value;
                OnPropertyChanged("Flash");
            }
        }

        public long UpdateTime
        {
            get
            {
                return _updateTime;
            }
            set
            {
                _updateTime = value;
                OnPropertyChanged("UpdateTime");
            }
        }

        public long AvgUpdateTime
        {
            get
            {
                return _avgUpdateTime;
            }
            set
            {
                _avgUpdateTime = value;
                OnPropertyChanged("AvgUpdateTime");
            }
        }

        public int QueueDepth
        {
            get
            {
                return _queueDepth;
            }
            set
            {
                _queueDepth = value;
                OnPropertyChanged("QueueDepth");
            }
        }

        public void IsSelected(bool isSelected)
        {
            _isSelected = isSelected;
        }

        private void MapControlView_Loaded(object sender, RoutedEventArgs e)
        {
            if (_isSelected && _isFirstLoad)
            {
                _isFirstLoad = false;

                try
                {
                    _mapId = MapId;

                    _workQueue = new ThreadedWorkQueue<WorkItem>("Updates");
                    _workQueue.OnProcessWorkItem += WorkQueue_OnProcessWorkItem;
                    _workQueue.StartUp();

                    mapView.MinimumScale = 0;
                    mapView.MapUnit = GeographyUnit.Meter;

                    _shapeLayer = new GraphicPointFeatureLayer();
                    _shapeLayer.Open();
                    _shapeLayer.Name = "ShapeLayer";
                    _shapeLayer.IsVisible = true;
                    MyFeatureLayers.Add(_shapeLayer);

                    _symbolLayer = new GraphicPointFeatureLayer();
                    _symbolLayer.Open();
                    _symbolLayer.Name = "SymbolLayer";
                    _symbolLayer.IsVisible = true;
                    MyFeatureLayers.Add(_symbolLayer);

                    //
                    // Choose the type of features to display
                    //
                    FeatureType featureType = FeatureType.PentagonShape;

                    if (featureType == FeatureType.PNGSymbol)
                    {
                        _featureWidth = 1.0;
                        _featureHeight = 1.0;
                        _distanceBetweenPointsX = 1.0;
                        _distanceBetweenPointsY = 1.0;
                        _zoomLevel20Scale = 200.0;
                        _zoomIncrement = 10.0;
                    }
                    else if (featureType == FeatureType.PentagonShape)
                    {
                        _featureWidth = 4.0;
                        _featureHeight = 4.0;
                        _distanceBetweenPointsX = 6.0;
                        _distanceBetweenPointsY = 6.0;
                        _zoomLevel20Scale = 200.0;
                        _zoomIncrement = 20.0;
                    }
                    else if (featureType == FeatureType.RectangleShape)
                    {
                        _featureWidth = 1.0;
                        _featureHeight = 1.0;
                        _distanceBetweenPointsX = 1.50;
                        _distanceBetweenPointsY = 0.5;
                        _zoomLevel20Scale = 100.0;
                        _zoomIncrement = 20.0;
                    }

                    

                    TotalPoints = _rowCount * _colCount;
                    Console.WriteLine("[Loaded] Total points - {0}", _totalPoints);

                    int featureCount = 1;
                    double minX = 0.0;
                    double maxY = 0.0;
                    double maxX = 0.0;
                    double minY = 0.0;

                    double x = minX;
                    double y = minY;

                    // Create features
                    for (int row = 0; row < _rowCount; row++)
                    {
                        for (int col = 0; col < _colCount; col++)
                        {
                            MyBaseFeature myFeature = null;

                            // Display features as PNG symbols
                            if (featureType == FeatureType.PNGSymbol)
                            {
                                myFeature = new MySymbolFeature(_mapId, x, y, col, row, _initialState);
                                _shapeLayer.Add(myFeature);
                            }
                            // Display features as a pentagon shape
                            else if (featureType == FeatureType.PentagonShape)
                            {
                                if ((featureCount % 2) == 0)
                                {
                                    myFeature = new MySymbolFeature(_mapId, x, y, col, row, _initialState);
                                    _symbolLayer.Add(myFeature);
                                }
                                else
                                {
                                    BaseShape myFeatureShape = CreatePentagonShape(x, y, _featureWidth, _featureHeight);
                                    myFeature = new MyShapeFeature(myFeatureShape, _mapId, x, y, col, row, _initialState);
                                    _shapeLayer.Add(myFeature);
                                }
                            }
                            // Display features as a rectangle shape
                            else if (featureType == FeatureType.RectangleShape)
                            {
                                BaseShape myFeatureShape = CreateRectangleShape(x, y, _featureWidth, _featureHeight);
                                myFeature = new MyShapeFeature(myFeatureShape, _mapId, x, y, col, row, _initialState);
                                _shapeLayer.Add(myFeature);
                            }

                            MainWindow.GlobalFeatureSource.AddGraphicPoint(myFeature);
                            x += _distanceBetweenPointsX;
                            featureCount++;

                            //if (_logDebug)
                            //    Console.WriteLine("[Window_Loaded] {0}: {1} X '{2}' Y '{3}'", myFeatureLayer.Name, myFeature.Id, myFeature.X, myFeature.Y);
                        }

                        maxX = x;
                        x = minX;
                        y += _distanceBetweenPointsY;
                    }

                    maxY = y;

                    NumFeaturesToUpdate = (int)((double)TotalPoints * 0.10);
                    //NumFeaturesToUpdate = (int)((double)TotalPoints * 0.008);
                    UpdateRate = 1000;

                    RectangleShape extent = new RectangleShape(minX, maxY, maxX, minY);

                    MyLineFeature lineFeature;

                    LineShape left = CreateLineShape(new PointShape(extent.LowerLeftPoint.X - _distanceBetweenPointsX, extent.LowerLeftPoint.Y), new PointShape(extent.UpperLeftPoint.X - _distanceBetweenPointsX, extent.UpperLeftPoint.Y));
                    lineFeature = new MyLineFeature(left, _mapId, extent.LowerLeftPoint.X, extent.UpperLeftPoint.Y / 2.0, (int)minX - 1, (int)maxY + 1, _initialState);
                    _shapeLayer.Add(lineFeature);
                    MainWindow.GlobalFeatureSource.AddGraphicPoint(lineFeature);
                    _lineFeatures.Add(lineFeature);

                    LineShape top = CreateLineShape(new PointShape(extent.UpperLeftPoint.X, extent.UpperLeftPoint.Y + _distanceBetweenPointsY), new PointShape(extent.UpperRightPoint.X, extent.UpperRightPoint.Y + _distanceBetweenPointsY));
                    lineFeature = new MyLineFeature(top, _mapId, extent.UpperRightPoint.X / 2.0, extent.UpperLeftPoint.Y, (int)maxY + 1, (int)minX - 1, _initialState);
                    _shapeLayer.Add(lineFeature);
                    MainWindow.GlobalFeatureSource.AddGraphicPoint(lineFeature);
                    _lineFeatures.Add(lineFeature);

                    LineShape right = CreateLineShape(new PointShape(extent.UpperRightPoint.X + _distanceBetweenPointsX, extent.UpperRightPoint.Y), new PointShape(extent.LowerRightPoint.X + _distanceBetweenPointsX, extent.LowerRightPoint.Y));
                    lineFeature = new MyLineFeature(right, _mapId, extent.UpperRightPoint.X, extent.UpperRightPoint.Y / 2.0, (int)maxX + 1, (int)minY - 1, _initialState);
                    _shapeLayer.Add(lineFeature);
                    MainWindow.GlobalFeatureSource.AddGraphicPoint(lineFeature);
                    _lineFeatures.Add(lineFeature);

                    LineShape bottom = CreateLineShape(new PointShape(extent.LowerLeftPoint.X, extent.LowerLeftPoint.Y - _distanceBetweenPointsY), new PointShape(extent.LowerRightPoint.X, extent.LowerRightPoint.Y - _distanceBetweenPointsY));
                    lineFeature = new MyLineFeature(bottom, _mapId, extent.LowerRightPoint.X / 2.0, extent.LowerRightPoint.Y, (int)minX - 1, (int)minY - 1, _initialState);
                    _shapeLayer.Add(lineFeature);
                    MainWindow.GlobalFeatureSource.AddGraphicPoint(lineFeature);
                    _lineFeatures.Add(lineFeature);

                    _graphicViewOverlay = new GraphicViewOverlay() { Name = _featureOverlayName };

                    // Create background image layer
                    if (_backgroundImage.Name != "None")
                    {
                        GraphicViewBackgroundData backgroundData = new GraphicViewBackgroundData(Guid.NewGuid(), _backgroundImage.Name, "desc", _backgroundImage.Filename, _backgroundImage.ImageBytes, "Black", GeographyUnit.Meter, minX - (maxX * 0.25), maxY * 1.25, maxX * 1.25, minY - (maxY * 0.25));
                        IGraphicBackgroundRasterImage backgroundRasterImage = GraphicBackgroundRasterImage.CreateFrom(backgroundData);
                        GraphicRasterLayer backgroundLayer = new GraphicRasterLayer(backgroundData, backgroundRasterImage);
                        _graphicViewOverlay.AddBackgroundLayer(backgroundLayer);
                    }

                    // Create the feature overlay and layers
                    _graphicViewOverlay.TileType = TileType.SingleTile;
                    _graphicViewOverlay.DrawingQuality = DrawingQuality.HighSpeed;
                    //_graphicViewOverlay.TransitionEffect = TransitionEffect.None;
                    _graphicViewOverlay.AddFeatureLayer(_shapeLayer);
                    _graphicViewOverlay.AddFeatureLayer(_symbolLayer);
                    _graphicViewOverlay.Drawing += _featureOverlay_Drawing;
                    _graphicViewOverlay.Drawn += _featureOverlay_Drawn;
                    _graphicViewOverlay.DrawingException += LayerOverlay_DrawingException;
                    mapView.Overlays.Add(_graphicViewOverlay);

                    OnPropertyChanged("FeatureLayers");

                    mapView.BackgroundOverlay.BackgroundBrush = new GeoSolidBrush(GeoColors.AliceBlue);
                    mapView.MapUnit = GeographyUnit.Meter;
                    mapView.ZoomLevelSet = new MyZoomLevelSet(_zoomLevel20Scale, _zoomIncrement);
                    mapView.CurrentExtent = extent;
                    mapView.MapTools.Logo.IsEnabled = false;
                    //mapView.MapTools.MouseCoordinate.IsEnabled = true;
                    //mapView.MapTools.PanZoomBar.IsEnabled = false;

                    mapView.EditOverlay = new MyEditOverlay();
                    mapView.TrackOverlay = new MyTrackOverlay();

                    mapView.Refresh();

                    _featureChangeTimer = new System.Timers.Timer();
                    _featureChangeTimer.AutoReset = true;
                    _featureChangeTimer.Interval = _featureChangeIntervalMS;
                    _featureChangeTimer.Elapsed += _featureChangeTimer_Elapsed;

                    _flashTimer = new FlashTimer();
                    _flashTimer.FlashTimerTick += _flashTimer_FlashTimerTick;                    
                }
                catch (Exception ex)
                {
                    Console.WriteLine("[Window_Loaded] Exception building map - {0}", ex.ToString());
                }
            }
        }

        private void MapControlView_Unloaded(object sender, RoutedEventArgs e)
        {

        }

        private void CalculateFeatureStates()
        {
            try
            {
                if (_logDebug)
                    Console.WriteLine();

                List<FeatureStatus> featuresToRefresh = new List<FeatureStatus>();

                if (_colNumber >= _colCount)
                {
                    _colNumber = 0;
                }

                if (_rowNumber >= _rowCount)
                {
                    _rowNumber = 0;
                }

                Console.WriteLine("{0} Calculate '{1}' feature states", DateTime.Now.ToString("HH:mm:ss,fff"), _numFeaturesToUpdate);

                lock (_lock)
                {
                    // Change the state of specified number of features
                    for (int i = 0; i < _numFeaturesToUpdate; i++)
                    {
                        if (Randomize)
                        {
                            _colNumber = _random.Next(0, _colCount);
                            _rowNumber = _random.Next(0, _rowCount);
                        }

                        MyBaseFeature myFeature = null;

                        if (_colNumber < _lineFeatures.Count && _rowNumber < _lineFeatures.Count)
                        {
                            if (_colNumber % 2 == 0)
                            {
                                myFeature = _lineFeatures[_colNumber];
                            }
                            else
                            {
                                myFeature = _lineFeatures[_rowNumber];
                            }
                        }
                        else
                        {
                            myFeature = GetFeature(_colNumber, _rowNumber);
                        }

                        if (myFeature == null)
                        {
                            Console.WriteLine("Unable to find feature at col '{0}' row '{1}", _colNumber, _rowNumber);
                            continue;
                        }


                        if (myFeature != null)
                        {
                            string featureState = myFeature.GetFeatureValue();
                            if (featureState == "0" || featureState == "00")
                                featuresToRefresh.Add(new FeatureStatus(myFeature.Id, "1", _rowNumber));
                            else if (featureState == "1" || featureState == "11")
                                featuresToRefresh.Add(new FeatureStatus(myFeature.Id, "2", _rowNumber));
                            else if (featureState == "2" || featureState == "22")
                                featuresToRefresh.Add(new FeatureStatus(myFeature.Id, "3", _rowNumber));
                            else if (featureState == "3" || featureState == "33")
                                featuresToRefresh.Add(new FeatureStatus(myFeature.Id, "0", _rowNumber));

                            if (++_colNumber >= _colCount)
                            {
                                _colNumber = 0;
                                if (++_rowNumber >= _rowCount)
                                {
                                    _rowNumber = 0;
                                }
                            }
                        }
                        else
                        {
                            Console.WriteLine("Unable to find layer for feature at ({0},{1})", _colNumber, _rowNumber);
                        }
                    }
                }

                WorkItem workItem = new WorkItem(true, featuresToRefresh);
                _workQueue.QueueWorkItem(workItem);

                int count = _workQueue.Count;
                QueueDepth = count;
                if (count > 0)
                {
                    Console.WriteLine("{0} Queue depth '{1}'", DateTime.Now.ToString("HH:mm:ss,fff"), count);
                }

                //_featureChangeTimer.Start();
            }
            catch (Exception ex)
            {
                Console.WriteLine("[CalculateFeatureStates] Exception updating map - {0}", ex.ToString());
            }
        }

        private void WorkQueue_OnProcessWorkItem(object sender, ThreadedWorkQueue<WorkItem>.ThreadedWorkQueueItemArgs e)
        {
            try
            {
                bool refresh = true;

                if (e.WorkItem.UpdateStatus)
                {
                    Stopwatch stopwatch = new Stopwatch();
                    stopwatch.Start();

                    lock (_lock)
                    {
                        Console.WriteLine("{0} Update    '{1}' features", DateTime.Now.ToString("HH:mm:ss,fff"), e.WorkItem.Updates.Count);

                        foreach (FeatureStatus featureStatus in e.WorkItem.Updates)
                        {
                            MyBaseFeature myFeature = GetFeature(featureStatus.Id);
                            if (myFeature == null)
                            {
                                Console.WriteLine("Unable to find feature id '{0}'", featureStatus.Id);
                                continue;
                            }

                            myFeature.SetFeatureValue(featureStatus.Status);
                            //Console.WriteLine("{0} Feature '{1}' status '{2}'", DateTime.Now.ToString("HH:mm:ss,fff"), myFeature.Id, myFeature.GetFeatureValue());

                            if (featureStatus.Status == "2" || featureStatus.Status == "22")
                            {
                                if (myFeature is MyShapeFeature)
                                {
                                    bool addFlashFeature = false;
                                    if (_shapeLayer.Contains(myFeature))
                                    {
                                        _shapeLayer.Remove(myFeature);
                                        if (_flashFeatures.Contains(myFeature))
                                        {
                                            addFlashFeature = true;
                                            _flashFeatures.Remove(myFeature);
                                        }
                                    }

                                    BaseShape myFeatureShape = CreatePentagonShape(myFeature.X, myFeature.Y, 2.0 * _featureWidth, 2.0 * _featureHeight);
                                    myFeature = new MyShapeFeature(myFeatureShape, _mapId, myFeature.X, myFeature.Y, myFeature.Col, myFeature.Row, featureStatus.Status);
                                    _shapeLayer.Add(myFeature);
                                    if (addFlashFeature)
                                    {
                                        _flashFeatures.Add(myFeature);
                                    }
                                }
                            }
                            else if (featureStatus.Status == "3" || featureStatus.Status == "33")
                            {
                                if (myFeature is MyShapeFeature)
                                {
                                    bool addFlashFeature = false;
                                    if (_shapeLayer.Contains(myFeature))
                                    {
                                        _shapeLayer.Remove(myFeature);
                                        if (_flashFeatures.Contains(myFeature))
                                        {
                                            addFlashFeature = true;
                                            _flashFeatures.Remove(myFeature);
                                        }
                                    }

                                    BaseShape myFeatureShape = CreatePentagonShape(myFeature.X, myFeature.Y, _featureWidth, _featureHeight);
                                    myFeature = new MyShapeFeature(myFeatureShape, _mapId, myFeature.X, myFeature.Y, myFeature.Col, myFeature.Row, featureStatus.Status);
                                    _shapeLayer.Add(myFeature);
                                    if (addFlashFeature)
                                    {
                                        _flashFeatures.Add(myFeature);
                                    }
                                }
                            }
                        }
                    }

                    stopwatch.Stop();
                    totalUpdateMS += stopwatch.ElapsedMilliseconds;
                    totalUpdateCount += 1;

                    UpdateTime = stopwatch.ElapsedMilliseconds;
                    AvgUpdateTime = totalUpdateMS / totalUpdateCount;
                }
                else if (e.WorkItem.Flash)
                {
                    if (Flash)
                    {
                        lock (_lock)
                        {
                            TimeSpan delta = DateTime.Now - _lastFlashTime;
                            _lastFlashTime = DateTime.Now;
                            Console.WriteLine("{0} Flash '{1}' symbols - Delta '{2}ms'", DateTime.Now.ToString("HH:mm:ss,fff"), _flashFeatures.Count, delta.TotalMilliseconds);

                            foreach (var feature in _flashFeatures)
                            {
                                string state = feature.GetFeatureValue();
                                if (e.WorkItem.FlashOn)
                                {
                                    if (state == "00")
                                        feature.SetFeatureValue("0");
                                    else if (state == "11")
                                        feature.SetFeatureValue("1");
                                    else if (state == "22")
                                        feature.SetFeatureValue("2");
                                    else if (state == "33")
                                        feature.SetFeatureValue("3");
                                }
                                else
                                {
                                    if (state == "0")
                                        feature.SetFeatureValue("00");
                                    else if (state == "1")
                                        feature.SetFeatureValue("11");
                                    else if (state == "2")
                                        feature.SetFeatureValue("22");
                                    else if (state == "3")
                                        feature.SetFeatureValue("33");
                                }
                            }
                        }
                    }
                    else
                    {
                        refresh = false;
                    }
                }

                if (refresh)
                {
                    Dispatcher.Invoke(new Action(() =>
                    {
                        Stopwatch stopwatch = new Stopwatch();
                        stopwatch.Start();

                        mapView.Refresh(_graphicViewOverlay);

                        stopwatch.Stop();
                        totalRefreshMS += stopwatch.ElapsedMilliseconds;
                        totalRefreshCount += 1;

                        refreshTime.Text = $"{stopwatch.ElapsedMilliseconds}";
                        totalRefreshTime.Text = $"{totalRefreshMS / totalRefreshCount}";
                    }));
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("[RefreshFeaturesThread] Exception - {0}", ex.ToString());
            }
        }

        private MyBaseFeature GetFeature(int col, int row)
        {
            string pointId = $"POINT({_mapId}){col}.{row}";
            MyBaseFeature myFeature = GetFeature(pointId);

            return myFeature;
        }

        private MyBaseFeature GetFeature(string id)
        {
            MyBaseFeature myFeature = (MyBaseFeature)_shapeLayer.GetFeatureById(id);
            if (myFeature == null)
            {
                myFeature = (MyBaseFeature)_symbolLayer.GetFeatureById(id);
            }

            return myFeature;
        }

        private void InitializeFlashFeatures()
        {
            _flashFeatures = new List<MyBaseFeature>();
            Random random = new Random();
            for (int i = 0; i < _numFeaturesToUpdate; i++)
            {
                int colNumber = random.Next(0, _colCount);
                int rowNumber = random.Next(0, _rowCount);
                MyBaseFeature feature = GetFeature(colNumber, rowNumber);
                if (feature != null)
                {
                    _flashFeatures.Add(feature);
                }
                else
                {
                    Console.WriteLine("Unable to find feature at col '{0}' row '{1}", colNumber, rowNumber);
                }
            }
        }

        private BaseShape CreatePentagonShape(double x, double y, double width, double height)
        {
            Collection<Vertex> vertices = new Collection<Vertex>();

            vertices.Add(new Vertex(x, y + (0.75 * height)));
            vertices.Add(new Vertex(x + (0.75 * width), y));
            vertices.Add(new Vertex(x + (0.50 * width), y - (0.75 * height)));
            vertices.Add(new Vertex(x - (0.50 * width), y - (0.75 * height)));
            vertices.Add(new Vertex(x - (0.75 * width), y));

            PolygonShape polygonShape = new PolygonShape(new RingShape(vertices));

            return polygonShape;
        }

        private BaseShape CreateRectangleShape(double x, double y, double width, double height)
        {
            Collection<Vertex> vertices = new Collection<Vertex>();

            vertices.Add(new Vertex(x + (0.5 * width), y + (0.25 * height)));
            vertices.Add(new Vertex(x + (0.5 * width), y - (0.25 * height)));
            vertices.Add(new Vertex(x - (0.50 * width), y - (0.25 * height)));
            vertices.Add(new Vertex(x - (0.50 * width), y + (0.25 * height)));

            PolygonShape polygonShape = new PolygonShape(new RingShape(vertices));

            return polygonShape;
        }

        private LineShape CreateLineShape(params PointShape[] points)
        {
            LineShape lineShape = new LineShape();
            foreach (PointShape ps in points)
            {
                lineShape.Vertices.Add(new Vertex(ps));
            }

            return lineShape;
        }

        private void RefreshMap(IEnumerable<RectangleShape> features)
        {
            _graphicViewOverlay.Refresh(features);
        }

        private void RefreshMap(RectangleShape feature)
        {
            _graphicViewOverlay.Refresh(feature);
        }

        private void LayerOverlay_DrawingException(object sender, DrawingExceptionTileOverlayEventArgs e)
        {
            Console.WriteLine("[LayerOverlay_DrawingException] Layer Overlay draw exception - {0}", e.Exception.Message);
        }

        private void _featureOverlay_Drawing(object sender, DrawingOverlayEventArgs e)
        {
            if (_logDebug)
                Console.WriteLine("[FeatureOverlay_Drawing] Start  draw feature overlay extent - {0}", e.WorldExtent.ToString());
        }

        private void _featureOverlay_Drawn(object sender, DrawnOverlayEventArgs e)
        {
            if (_logDebug)
                Console.WriteLine("[FeatureOverlay_Drawn] Finish draw feature overlay extent - {0}", e.WorldExtent.ToString());
        }

        private void mapView_MouseMove(object sender, MouseEventArgs e)
        {
            //Gets the PointShape in world coordinates from screen coordinates.
            System.Windows.Point point = e.MouseDevice.GetPosition(null);

            ScreenPointF screenPointF = new ScreenPointF((float)point.X, (float)point.Y);
            PointShape pointShape = MapUtil.ToWorldCoordinate(mapView.CurrentExtent, screenPointF, (float)mapView.ActualWidth, (float)mapView.ActualHeight);

            textBox1.Text = "X: " + pointShape.X + "  Y: " + pointShape.Y;
        }

        private void _featureChangeTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            CalculateFeatureStates();
        }

        private void _flashTimer_FlashTimerTick(object sender, FlashTimerTickEventPayload e)
        {
            WorkItem workItem = new WorkItem(true, e.FlashCommand == FlashCommandEnum.ON);
            _workQueue.QueueWorkItem(workItem);

            int count = _workQueue.Count;
            QueueDepth = count;
            if (count > 0)
            {
                Console.WriteLine("{0} Queue depth '{1}'", DateTime.Now.ToString("HH:mm:ss,fff"), count);
            }
        }

        private void StartButton_Click(object sender, RoutedEventArgs e)
        {
            _random = new Random();
            _colNumber = 0;
            _rowNumber = 0;
            _featureChangeTimer.Interval = _featureChangeIntervalMS;
            _featureChangeTimer.Start();
            _flashTimer.Start();
            _lastFlashTime = DateTime.Now;
            InitializeFlashFeatures();
        }

        private void StopButton_Click(object sender, RoutedEventArgs e)
        {
            _featureChangeTimer.Stop();
            //_flashTimer.Stop();
        }

        private void BeginEditButton_Click(object sender, RoutedEventArgs e)
        {
            foreach (Feature feature in _shapeLayer.GetAllGraphicPoints())
            {
                mapView.EditOverlay.EditShapesLayer.InternalFeatures.Add(feature.Id, feature);
            }

            _shapeLayer.ClearAll();


            foreach (Feature feature in _symbolLayer.GetAllGraphicPoints())
            {
                mapView.EditOverlay.EditShapesLayer.InternalFeatures.Add(feature.Id, feature);
            }

            _symbolLayer.ClearAll();

            mapView.EditOverlay.CalculateAllControlPoints();

            mapView.Refresh(new Overlay[] { mapView.EditOverlay, _graphicViewOverlay });
        }

        private void EndEditButton_Click(object sender, RoutedEventArgs e)
        {
            foreach (Feature feature in mapView.EditOverlay.EditShapesLayer.InternalFeatures)
            {
                if (feature is MySymbolFeature)
                {
                    _symbolLayer.Add((MySymbolFeature)feature);
                }
                else if (feature is MyShapeFeature)
                {
                    _shapeLayer.Add((MyShapeFeature)feature);
                }
                else
                {
                    Console.WriteLine("Unknown feature type '{0}'", feature.GetType().Name);
                }
            }

            mapView.EditOverlay.EditShapesLayer.InternalFeatures.Clear();

            mapView.Refresh(new Overlay[] { mapView.EditOverlay, _graphicViewOverlay });
        }

        private void ResetButton_Click(object sender, RoutedEventArgs e)
        {
            totalRefreshMS = 0;
            totalRefreshCount = 0;
            refreshTime.Text = string.Empty;
            totalRefreshTime.Text = string.Empty;
            UpdateTime = 0;
            AvgUpdateTime = 0;
            QueueDepth = 0;

            _workQueue.ClearAll();
            _featureChangeTimer.Stop();
            foreach (var layer in _featureLayers)
            {
                foreach (var feature in layer.FeatureSource.GetAllFeatures(ReturningColumnsType.NoColumns))
                {
                    MyBaseFeature myBaseFeature = feature as MyBaseFeature;
                    myBaseFeature.SetFeatureValue("0");
                }
            }

            _graphicViewOverlay.Refresh();
        }


        private void OnPropertyChanged(string name)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(name));
        }
    }

    public class MyZoomLevelSet : ZoomLevelSet
    {
        public MyZoomLevelSet(double start, double offset)
        {
            ZoomLevel20.Scale = start;
            ZoomLevel19.Scale = start += offset;
            ZoomLevel18.Scale = start += offset;
            ZoomLevel17.Scale = start += offset;
            ZoomLevel16.Scale = start += offset;
            ZoomLevel15.Scale = start += offset;
            ZoomLevel14.Scale = start += offset;
            ZoomLevel13.Scale = start += offset;
            ZoomLevel12.Scale = start += offset;
            ZoomLevel11.Scale = start += offset;
            ZoomLevel10.Scale = start += offset;
            ZoomLevel09.Scale = start += offset;
            ZoomLevel08.Scale = start += offset;
            ZoomLevel07.Scale = start += offset;
            ZoomLevel06.Scale = start += offset;
            ZoomLevel05.Scale = start += offset;
            ZoomLevel04.Scale = start += offset;
            ZoomLevel03.Scale = start += offset;
            ZoomLevel02.Scale = start += offset;
            ZoomLevel01.Scale = start += offset;

            Console.WriteLine("[MyZoomLevelSet] Zoom level 20 scale is '{0}'", ZoomLevel20.Scale);
            Console.WriteLine("[MyZoomLevelSet] Zoom level 1 scale is '{0}'", ZoomLevel01.Scale);
        }
    }

    public class WorkItem
    {
        public bool UpdateStatus { get; set; }
        public List<FeatureStatus> Updates { get; set; }
        public bool Flash { get; set; }
        public bool FlashOn { get; set; }

        public WorkItem(bool updateStatus, List<FeatureStatus> updates)
        {
            UpdateStatus = updateStatus;
            Updates = updates;
        }

        public WorkItem(bool flash, bool flashOn)
        {
            Flash = flash;
            FlashOn = flashOn;
        }
    }

    public class FeatureStatus
    {
        public string Id { get; set; }
        public string Status { get; set; }
        public int Layer { get; set; }

        public FeatureStatus(string id, string status, int layer)
        {
            Id = id;
            Status = status;
            Layer = layer;
        }
    }
}
