﻿using System;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Arinc.UI.Common.Utility
{
    public class MemoryOptimizer
    {
        #region Members

        private bool _isEnabled;
        private int _optimizationIntervalInSeconds;
        private int _optimizationThresholdInMB;
        private System.Timers.Timer _processTimer;

        [DllImport("KERNEL32.DLL", EntryPoint = "SetProcessWorkingSetSize", SetLastError = true, CallingConvention = CallingConvention.StdCall)]
        static extern bool SetProcessWorkingSetSize32(IntPtr pProcess, int dwMinimumWorkingSetSize, int dwMaximumWorkingSetSize);

        [DllImport("KERNEL32.DLL", EntryPoint = "SetProcessWorkingSetSize", SetLastError = true, CallingConvention = CallingConvention.StdCall)]
        static extern bool SetProcessWorkingSetSize64(IntPtr pProcess, long dwMinimumWorkingSetSize, long dwMaximumWorkingSetSize);

        [DllImport("psapi.dll")]
        static extern int EmptyWorkingSet(IntPtr hwProc);

        #endregion // Members

        #region Constructors
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="logger"></param>
        public MemoryOptimizer(int optimizationIntervalInSeconds, int optimizationThresholdInMB, bool isEnabled)
        {
            _optimizationIntervalInSeconds = optimizationIntervalInSeconds;
            _optimizationThresholdInMB = optimizationThresholdInMB;
            _isEnabled = isEnabled;

            InitializeProcessTimer();
        }

        #endregion // Constructors

        #region Events
        #endregion // Events

        #region Properties

        #region Public Properties

        public bool IsEnabled
        {
            get
            {
                return _isEnabled;
            }
            set
            {
                _isEnabled = value;
            }
        }

        /// <summary>
        /// Gets or Sets the thread execution time interval parameter.
        /// </summary>
        public int OptimizationIntervalInSeconds
        {
            get
            {
                return _optimizationIntervalInSeconds;
            }
            set
            {
                _optimizationIntervalInSeconds = value;

                if (_processTimer != null)
                {
                    _processTimer.Stop();
                    _processTimer.Interval = _optimizationIntervalInSeconds * 1000;
                    _processTimer.Start();
                }
            }
        }

        /// <summary>
        /// Gets or Sets the application threshold memory parameter.
        /// </summary>
        public int OptimizationThresholdInMB
        {
            get
            {
                return _optimizationThresholdInMB;
            }
            set
            {
                _optimizationThresholdInMB = value;
            }
        }

        #endregion // Public Properties

        #region Protected Properties
        #endregion // Protected Properties

        #region Private Properties
        #endregion // Private Properties

        #endregion // Properties

        #region Methods

        #region Public Methods
        /// <summary>
        /// Method to compact the large object heap.
        /// By default the large object heap are not compacted which can lead to out of memory exception.
        /// Observed that - Badges (30k records) and Authorization is been allocated in LOH, and never get compacted
        /// Forcing GC to compact the LOH.
        /// https://www.simple-talk.com/dotnet/.net-framework/the-dangers-of-the-large-object-heap/
        /// </summary>
        public static void ForceMemoryOptimizationAndLOHCompaction(bool compachLOH = false)
        {
            FlushApplicationMemory(compachLOH);
        }

        #endregion // Public Methods

        #region Protected Methods
        #endregion // Protected Methods

        #region Private Methods
        /// <summary>
        /// Method to initialize process timer.
        /// </summary>
        private void InitializeProcessTimer()
        {
            _processTimer = new System.Timers.Timer();
            _processTimer.AutoReset = false;
            _processTimer.Interval = OptimizationIntervalInSeconds * 1000;
            _processTimer.Elapsed += ProcessTimerElapsed;
            _processTimer.Start();
        }

        /// <summary>
        /// Process timer Elapsed handler.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ProcessTimerElapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if (IsEnabled)
            {
                PerformMemoryOptimization();
            }

            _processTimer.Start();
        }

        /// <summary>
        /// Monitored thread delegate method.
        /// </summary>
        private void PerformMemoryOptimization()
        {
            var beforeFlushMemoryInBytes = GC.GetTotalMemory(false);

            long threshold = OptimizationThresholdInMB * 100000;

            Console.WriteLine("Current memory usage: '{0}'  Threshold: '{1}'", beforeFlushMemoryInBytes, threshold);

            // check wether application need force clean up or not, if not rely upon GC to kick off by framework.
            // else initiate force clean up all generation
            if (beforeFlushMemoryInBytes >= threshold)
            {
                Console.WriteLine("Preparing to explicit flushing memory, Total memory Utilization is {0} bytes.", beforeFlushMemoryInBytes);

                FlushApplicationMemory();

                beforeFlushMemoryInBytes = GC.GetTotalMemory(true);

                Console.WriteLine("Bought down memory to : {0} bytes.", beforeFlushMemoryInBytes);

                Thread.Sleep(3000);
            }
        }

        /// <summary>
        /// Method to force application to free it's memory and working set.
        /// More Info : Mysteries of Windows Memory Management Revealed part 1 & part 2
        /// https://blogs.msdn.microsoft.com/tims/2010/10/28/pdc10-mysteries-of-windows-memory-management-revealed-part-one/
        /// https://blogs.msdn.microsoft.com/tims/2010/10/29/pdc10-mysteries-of-windows-memory-management-revealed-part-two/
        /// </summary>
        private static void FlushApplicationMemory(bool compachLOH = true)
        {
            // Forcing GC to clean up the Large generations i.e. Gen2 only & compact the large object heap.

            if (compachLOH)
                System.Runtime.GCSettings.LargeObjectHeapCompactionMode = System.Runtime.GCLargeObjectHeapCompactionMode.CompactOnce;

            GC.Collect(2, GCCollectionMode.Forced);
            GC.WaitForPendingFinalizers();

            if (Environment.OSVersion.Platform == PlatformID.Win32NT)
            {
                using (Process process = Process.GetCurrentProcess())
                {
                    if (IntPtr.Size == 8)
                        SetProcessWorkingSetSize64(process.Handle, -1, -1);
                    else
                        SetProcessWorkingSetSize32(process.Handle, -1, -1);

                    // When the application is idle or the the memory runs low the OS will eventually reclaim / reduce the working set of an application. 
                    // So, calling the EmptyWorkingSet / SetProcessWorkingSetSize is just a way to clean up before the OS does it.
                    // forcing to empty as much as possible of its working set.
                    EmptyWorkingSet(process.Handle);
                }
            }
        }

        #endregion // Private Methods

        #endregion // Methods
    }
}
