﻿/***************************************************************************
*                               COPYRIGHT                                    
*                   (c) 2014 ARINC, All rights reserved.                     
*                                                                            
* RESTRICTED RIGHTS LEGEND                                                   
* This  program is  the licensed  property of  ARINC,  and/or  its licensors 
* and  its use  is restricted  by license to  the computer  system  hardware 
* described in  the applicable  purchase document.   Unless ARINC, Inc.  has 
* provided  its express  written consent, the  contents of the  program  may 
* not  be  disclosed to  any third party  or used for any purpose other than 
* purpose  for  which  it  is  licensed  by  ARINC.   This program  is  also 
* protected  by international  and  U.S. copyright  laws.  Any  unauthorized 
* reproduction  or publication  is prohibited.   If the licensed user is any 
* unit  or agency  of the  United States  Government, then the licensed user 
* agrees that:  (a) the  program is  "Commercial Computer Software"  as  the 
* term is  defined in  Paragraph 27.471 of the DoD supplement to the Federal 
* Acquisitions   Regulations  and  in   any  corresponding   DAR  and   ASPR 
* Regulations;  (b)  the  program  is  developed  at private expense, and no 
* part  of it was  developed  with government  funds;  (c) the  government's 
* use of  the program  is subject  to "Restricted  Rights" as  that  term is 
* defined  in  Subdivision  (c)(1)(ii)  of  the  DoD  FAR-Supp  52.227-7013; 
* (d)   the  program   is  a  trade  secret  of  ARINC   for  all   purposes 
* including  for the  purpose of  the Freedom of  Information Act;  (e) each 
* copy  of the  program supplied  to the government or to another government 
* contractor  will, also bear  and/or have  affixed near it  the  restricted 
* rights legend.                                                             
***************************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Diagnostics;

namespace ThinkGeoPerformanceApp
{
    public interface IThreadWorkQueue<T>
    {
        string Name { get; set; }

        /// <summary>
        /// Built-in highest priority
        /// </summary>
        int HIGHEST_PRIORITY { get; }

        /// <summary>
        /// Built-in high priority
        /// </summary>
        int HIGH_PRIORITY { get; }

        /// <summary>
        /// Built-in medium priority
        /// </summary>
        int MEDIUM_PRIORITY { get; }

        /// <summary>
        /// Built-in low priority
        /// </summary>
        int LOW_PRIORITY { get; }

        /// <summary>
        /// Built-in lowest priority
        /// </summary>
        int LOWEST_PRIORITY { get; }

        event ThreadedWorkQueue<T>.OnWorkItemDelegate OnProcessWorkItem;
        SortedDictionary<int, List<T>> GetWorkQueueItems();
        void QueueWorkItem(T workItem);
        void QueueWorkItem(T workItem, int priority);
        void ClearAll();
        void Pause();
        void Resume();
        void StartUp();
        void Shutdown();
        void ShutdownNow();
    }

    /// <summary>
    /// This class implements a threaded (and thread-safe) priority FIFO work queue. Work items
    /// are delivered sequentially on a separate thread via the OnProcessWorkItem event. Higher priority
    /// work items (larger value) are delivered before lower priority work items (lower value). 
    /// For example, priority 10 is higher than priority 1.
    /// </summary>
    /// <typeparam name="T">The type of object to queue</typeparam>
    ///     
    public class ThreadedWorkQueue<T> : IThreadWorkQueue<T>, IDisposable
    {
        #region Members

        private string _name;
        private SortedDictionary<int/*priority*/, Queue<_WorkQueueItem>> _workQueues = new SortedDictionary<int/*priority*/, Queue<_WorkQueueItem>>(new PriorityDecendingComparer());
        private Thread _workerThread = null;
        private bool _bStopWorkerThreadDontUseMe = false;
        private object _stopWorkerThreadLock = new object();
        private bool _bPaused = false;
        private object _processLock = new object();
        private AutoResetEvent _itemQueuedAutoResetEvent = new AutoResetEvent(false);
        private bool _isDisposed = false;

        /// <summary>
        /// Built-in highest priority
        /// </summary>
        public int HIGHEST_PRIORITY { get { return 20; } }

        /// <summary>
        /// Built-in high priority
        /// </summary>
        public int HIGH_PRIORITY { get { return 15; } }

        /// <summary>
        /// Built-in medium priority
        /// </summary>
        public int MEDIUM_PRIORITY { get { return 10; } }

        /// <summary>
        /// Built-in low priority
        /// </summary>
        public int LOW_PRIORITY { get { return 5; } }

        /// <summary>
        /// Built-in lowest priority
        /// </summary>
        public int LOWEST_PRIORITY { get { return 0; } }

        #endregion

        #region Events

        /// <summary>
        /// Delegate used to specify if debug logging is enabled for 
        /// work item processing. Allows users of this class to tailor
        /// the debug logging associated with queueing and delivering
        /// a work item. For example, this delegate can be implemented 
        /// to return the IsFineDebugEnabled state of the user's logger.
        /// In this way, work item processing will only be logged if the
        /// user's logger has fine debug enabled. Note that this does NOT 
        /// override the Debug enabled status of the ThreadedWorkQueue 
        /// component in LogManager.xml. If Debug is disabled for 
        /// ThreadedWorkQueue in LogManager.xml then all ThreadedWorkQueue
        /// debug is disabled regardless of the value returned by this 
        /// delegate.
        /// </summary>
        /// <returns></returns>
        public delegate bool LogWorkItemProcessingDelegate();

        /// <summary>
        /// Delegate used to process a work item
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void OnWorkItemDelegate(object sender, ThreadedWorkQueueItemArgs e);

        /// <summary>
        /// Event delivered for each work item on the queue
        /// </summary>
        public event OnWorkItemDelegate OnProcessWorkItem;

        /// <summary>
        /// Work item event argument
        /// </summary>
        public class ThreadedWorkQueueItemArgs : EventArgs
        {
            /// <summary>
            /// The work item to process
            /// </summary>
            public T WorkItem { get; private set; }

            /// <summary>
            /// The priority of the work item. The higher the 
            /// priority the sooner the work item is processed.
            /// </summary>
            public int Priority { get; private set; }

            /// <summary>
            /// If true, then the work item will be removed from queue after processing.
            /// Default value is 'true'. If false, then leave the work item on the queue
            /// and automatically pause the queue. The queue must be resumed before processing
            /// will continue.
            /// </summary>
            public bool RemoveWorkItem { get; set; }

            /// <summary>
            /// Constructor
            /// </summary>
            /// <param name="workItem"></param>
            /// <param name="priority"></param>
            public ThreadedWorkQueueItemArgs(T workItem, int priority)
            {
                WorkItem = workItem;
                Priority = priority;

                RemoveWorkItem = true;
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Name used to identify this Threaded Queue in log messages
        /// </summary>
        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

        /// <summary>
        /// Gets the number of items in the work queue
        /// </summary>
        public int Count
        {
            get
            {
                int count = 0;
                lock (_workQueues)
                {
                    foreach (Queue<_WorkQueueItem> queue in _workQueues.Values)
                    {
                        count += queue.Count;
                    }
                }

                return count;
            }
        }

        /// <summary>
        /// Gets/Sets a flag indicating if the worker thread is to exit
        /// </summary>
        private bool _bStopWorkerThread
        {
            get
            {
                bool stop;
                lock (_stopWorkerThreadLock)
                {
                    stop = _bStopWorkerThreadDontUseMe;
                }

                return stop;
            }
            set
            {
                lock (_stopWorkerThreadLock)
                {
                    _bStopWorkerThreadDontUseMe = value;
                }
            }
        }

        /// <summary>
        /// Delegate used to specify if debug logging is enabled for 
        /// work item processing. Allows users of this class to tailor
        /// the debug logging associated with queueing and delivering
        /// a work item. For example, this delegate can be implemented 
        /// to return the IsFineDebugEnabled state of the user's logger.
        /// In this way, work item processing will only be logged if the
        /// user's logger has fine debug enabled. Note that this does NOT 
        /// override the Debug enabled status of the ThreadedWorkQueue 
        /// component in LogManager.xml. If Debug is disabled for 
        /// ThreadedWorkQueue in LogManager.xml then all ThreadedWorkQueue
        /// debug is disabled regardless of the value returned by this 
        /// delegate.
        /// </summary>
        /// <returns></returns>
        public LogWorkItemProcessingDelegate LogWorkItemProcessing { get; set; }

        #endregion

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="name">A name to use for logging purposes</param>
        public ThreadedWorkQueue(string name)
        {
            _name = name;
            LogWorkItemProcessing = null;
        }
        public ThreadedWorkQueue()
        {
            LogWorkItemProcessing = null;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Returns a list of all the items currently in the queue, sorted decending by priority. Higher priority
        /// work items appear higher in the dictionary.
        /// </summary>
        /// <returns></returns>
        public SortedDictionary<int, List<T>> GetWorkQueueItems()
        {
            SortedDictionary<int, List<T>> items = new SortedDictionary<int, List<T>>();

            lock (_workQueues)
            {
                foreach (int priority in _workQueues.Keys)
                {
                    List<T> priorityItems = new List<T>();
                    items[priority] = priorityItems;
                    foreach (_WorkQueueItem wqi in _workQueues[priority])
                    {
                        priorityItems.Add(wqi.WorkItem);
                    }
                }
            }

            return items;
        }

        /// <summary>
        /// Queues a work item for processing at the medium priority in the _run thread
        /// </summary>
        /// <param name="workItem">Work item to be queued</param>
        public void QueueWorkItem(T workItem)
        {
            QueueWorkItem(workItem, MEDIUM_PRIORITY);
        }

        /// <summary>
        /// Queues a work item for processing at the specified priority.
        /// </summary>
        /// <param name="workItem">Work item to be queued.</param>
        /// <param name="priority">Priority of the work item. The smaller the value the higher the priority.</param>
        public void QueueWorkItem(T workItem, int priority)
        {
            _queueWorkItem(new ThreadedWorkQueue<T>._WorkQueueItem(workItem, priority));
        }

        /// <summary>
        /// Queues a work item for processing at the specified priority.
        /// </summary>
        /// <param name="workItem">Work item to be queued.</param>
        private void _queueWorkItem(_WorkQueueItem workItem)
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>._queueWorkItem]", _name);

            try
            {
                if (!_isDisposed)
                {
                    if (workItem != null)
                    {
                        lock (_workQueues)
                        {
                            // Add a queue for the specified priority if it does
                            // not already exist
                            Queue<_WorkQueueItem> queue;
                            if (!_workQueues.TryGetValue(workItem.Priority, out queue))
                            {
                                queue = new Queue<_WorkQueueItem>();
                                _workQueues.Add(workItem.Priority, queue);
                            }

                            // Add the work item to its priority queue
                            if (_isLogWorkItemProcessingEnabled())
                                _logFineDebug(methodName, "Queue work item - {0}", workItem.ToString());
                            queue.Enqueue(workItem);
                        }

                        // Let the processing thread know it has a work item to process
                        _itemQueuedAutoResetEvent.Set();
                    }
                    else
                    {
                        _logError(methodName, "Cannot queue '<null>' work item");
                        throw new ArgumentNullException("workItem");
                    }
                }
                else
                {
                    _logInformation(methodName, "Not queuing work item - {0} - The work queue is disposed", workItem.ToString());
                }
            }
            catch (Exception ex)
            {
                _logError(methodName, "Error queuing work item - {0}", ex.ToString());
            }
        }

        /// <summary>
        /// Clears all work items in all priority queues
        /// </summary>
        public void ClearAll()
        {
            string methodName = "[ThreadedWorkQueue.ClearAll]";

            try
            {
                _logFineDebug(methodName, "Begin  clear all work queue items...");
                
                lock (_workQueues)
                {
                    _workQueues.Clear();
                }

                _logFineDebug(methodName, "Finish clear all work queue items");
            }
            catch (Exception ex)
            {
                _logError(methodName, "Exception - {0}", ex.ToString());
            }
        }

        /// <summary>
        /// Pauses processing of the work queue. Items can be added to the queue,
        /// but they will not be delivered until the queue has been resumed.
        /// </summary>
        public void Pause()
        {
            lock (_processLock)
            {
                _bPaused = true;
            }
        }

        /// <summary>
        /// Resumes processing of the work queue.
        /// </summary>
        public void Resume()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>.Resume]", _name);

            _logFineDebug(methodName, "Begin  resume work queue...");

            lock (_processLock)
            {
                lock (_workQueues)
                {                    
                    _bPaused = false;

                    // If there are work items queued then process 
                    // them immediately after being resumed
                    if (_workQueues.Count > 0)
                    {
                        _logFineDebug(methodName, "Work queue has '{0}' work items on resume - Process queued work items", _workQueues.Count);
                        _itemQueuedAutoResetEvent.Set();
                    }
                }
            }

            _logFineDebug(methodName, "Finish resume work queue");
        }

        /// <summary>
        /// Worker thread. All requests are queued and then delivered on this thread.
        /// </summary>
        private void _run()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>._run]", _name);

            _logFineDebug(methodName, "Threaded work queue running");

            while (true)
            {
                // Check if we are to exit immediately
                if (_bStopWorkerThread)
                {
                    _logFineDebug(methodName, "Stopping the threaded work queue immediately...");
                    break;
                }

                try
                {
                    // Wait for a work item to be queued
                    if (_itemQueuedAutoResetEvent.WaitOne(10000))
                    {
                        _logFineDebug(methodName, "Processing queued items");
                    }
                    else
                    {
                        if (_isLogWorkItemProcessingEnabled())
                            _logFineDebug(methodName, "Perform periodic queue processing");
                    }

                    // Deliver all items currently in the queue
                    while (true)
                    {
                        // Check if we are to exit immediately
                        if (_bStopWorkerThread)
                        {
                            _logFineDebug(methodName, "Received 'immediate' request to stop the threaded work queue...");
                            break;
                        }

                        lock (_processLock)
                        {
                            if (!_bPaused)
                            {


                                // Only process X number of work items for a priority before going
                                // to the next lowest priority


                                // Get the next work item on the queue. Note that this does NOT
                                // remove the work item from the queue - it just peeks. The subscriber
                                // must tell us that it is OK to remove the work item from the queue.
                                _WorkQueueItem workItem = _getWorkItem();
                                if (workItem != null)
                                {
                                    if (!workItem.Exit)
                                    {
                                        ThreadedWorkQueueItemArgs response = _notifySubscribersOfWorkItem(workItem);
                                        if (response != null)
                                        {
                                            _processSubscriberResponse(response);
                                        }
                                        else
                                        {
                                            break;
                                        }
                                    }
                                    else
                                    {
                                        _logFineDebug(methodName, "Received 'queued' request to stop the threaded work queue...");

                                        // _getWorkItem() does not remove the work item from the queue - it just peeks
                                        // So we need to explicitly remove the exit work item
                                        _removeWorkItem(workItem);

                                        // Signal that the worker thread should exit
                                        _bStopWorkerThread = true;

                                        break;
                                    }
                                }
                                else
                                {
                                    break;
                                }
                            }
                            else
                            {
                                break;
                            }
                        }
                    }
                }
                catch (ThreadAbortException)
                {
                    _logDebug(methodName, "Received thread abort request...");
                }
                catch (Exception ex)
                {
                    _logError(methodName, "Error in worker thread - {0}", ex.ToString());
                }
            }

            // Reset latch used to stop the worker thread so it won't immediately
            // stop the next time it is run.
            _bStopWorkerThread = false;

            _logFineDebug(methodName, "Threaded work queue has stopped");
        }

        /// <summary>
        /// Starts the worker thread used to deliver work event items
        /// </summary>
        public void StartUp()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>.StartUp]", _name);

            // Don't start our worker thread if it is already running
            if (_workerThread == null || !_workerThread.IsAlive)
            {
                // Start our worker thread
                _logFineDebug(methodName, "Starting worker thread...");
                _workerThread = new Thread(new ThreadStart(_run));
                _workerThread.IsBackground = true;
                _workerThread.Name = string.Format("{0} TWQ:{1}", _workerThread.ManagedThreadId, _name);
                _workerThread.Start();
                _logFineDebug(methodName, "Started worker thread");
            }
            else
            {
                _logDebug(methodName, "NOT starting worker thread - The worker thread is already running");
            }
        }

        /// <summary>
        /// Shuts down the work queue. Note that a request is queued at the highest
        /// possible priority (int.MaxValue) to shutdown the work queue.
        /// That is, all work items currently queued at the highest priority will be
        /// processed BEFORE the work queue actually shutsdown. To shutdown the work
        /// queue immediately call ShutdownNow(). Note that the work queue is NOT
        /// cleared as a result of the shutdown. Call ClearAll() to remove all items
        /// from the work queue.
        /// </summary>
        public void Shutdown()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>.Shutdown]", _name);

            if (_workerThread != null)
            {
                try
                {
                    // Don't shutdown the worker thread if it is not running
                    if (_workerThread != null && _workerThread.IsAlive)
                    {
                        _logFineDebug(methodName, "Queueing request to stop worker thread");

                        // Queue a request for the worker thread to exit
                        T dummy = default(T);
                        _WorkQueueItem qi = new ThreadedWorkQueue<T>._WorkQueueItem(dummy, int.MaxValue);
                        qi.Exit = true;
                        _queueWorkItem(qi);
                    }
                    else
                    {
                        _logDebug(methodName, "NOT shutting down worker thread - The worker thread is not running");
                    }
                }
                catch (Exception ex)
                {
                    _logError(methodName, "Error stopping worker thread - {0}", ex.ToString());
                }
            }
        }

        /// <summary>
        /// Shuts down the work queue immediately. Note that the work queue is NOT
        /// cleared as a result of the shutdown. Call ClearAll() to remove all items
        /// from the work queue.
        /// </summary>
        public void ShutdownNow()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>.ShutdownNow]", _name);

            try
            {
                _logFineDebug(methodName, "Received request to shutdown work queue immediately");
                _bStopWorkerThread = true;
                _itemQueuedAutoResetEvent.Set();
            }
            catch (Exception ex)
            {
                _logError(methodName, "Exception - {0}", ex.ToString());
            }
        }

        /// <summary>
        /// Notifies subscribers to process a work item.
        /// </summary>
        /// <param name="workItem"></param>
        /// <returns>Returns the response to the notification. Returns null if no subscribers to notify.</returns>
        private ThreadedWorkQueueItemArgs _notifySubscribersOfWorkItem(_WorkQueueItem workItem)
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>._notifySubscribersOfWorkItem]", _name);

            ThreadedWorkQueueItemArgs args = null;
            
            try
            {
                if (!_isDisposed)
                {
                    if (workItem != null && OnProcessWorkItem != null)
                    {
                        if (_isLogWorkItemProcessingEnabled())
                            _logDebug(methodName, "Process work item - [Counts {0}] - {1}", _getWorkQueueItemCount(), workItem.ToString());
                        args = new ThreadedWorkQueueItemArgs(workItem.WorkItem, workItem.Priority);
                        OnProcessWorkItem(this, args);
                    }
                }
                else
                {
                    _logInformation(methodName, "Not delivering work item - {0} - The work queue is disposed", workItem.ToString());
                }
            }
            catch (Exception ex)
            {
                _logError(methodName, "Error processing work item - {0}", ex.ToString());
            }

            return args;
        }

        /// <summary>
        /// Processes the response returned by the work item owner
        /// </summary>
        /// <param name="response"></param>
        private void _processSubscriberResponse(ThreadedWorkQueueItemArgs response)
        {
            if (response.RemoveWorkItem)
            {
                _removeWorkItem(new _WorkQueueItem(response.WorkItem, response.Priority));
            }
            else
            {
                Pause();
            }
        }

        /// <summary>
        /// Returns the oldest highest priority entry in the work queue. Returns null if the
        /// work queue is empty.
        /// </summary>
        /// <returns></returns>
        private _WorkQueueItem _getWorkItem()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>._getWorkItem]", _name);

            _WorkQueueItem workItem = null;

            try
            {
                // Get the highest priority work item on the queue
                lock (_workQueues)
                {
                    foreach (int priority in _workQueues.Keys)
                    {
                        Queue<_WorkQueueItem> queue = _workQueues[priority];
                        if (queue.Count > 0)
                        {
                            _logWorkQueueItemCount();
                            workItem = queue.Peek();
                            break;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                _logError(methodName, "Error getting oldest work item - {0}", ex.ToString());
            }

            return workItem;
        }

        /// <summary>
        /// Removes the work item at the head of the specified priority queue
        /// </summary>
        /// <param name="workItem"></param>
        private void _removeWorkItem(_WorkQueueItem workItem)
        {
            lock (_workQueues)
            {
                if (_workQueues.ContainsKey(workItem.Priority))
                {
                    _workQueues[workItem.Priority].Dequeue();
                }
            }
        }

        /// <summary>
        /// Returns a string that contains the number of entries in
        /// each queue.
        /// </summary>
        /// <returns></returns>
        private string _getWorkQueueItemCount()
        {
            StringBuilder counts = new StringBuilder();

            lock (_workQueues)
            {
                foreach (int priority in _workQueues.Keys)
                {
                    if (counts.Length > 0) counts.Append(" ");
                    counts.Append(string.Format("Priority[{0}] = '{1}'", priority, _workQueues[priority].Count));
                }
            }

            return counts.ToString();
        }

        /// <summary>
        /// Logs the count of work items by priority
        /// </summary>
        private void _logWorkQueueItemCount()
        {
            if (_isLogWorkItemProcessingEnabled())
            {
                string methodName = string.Format("[ThreadedWorkQueue<{0}>._logWorkQueueItemCount]", _name);

                _logFineDebug(methodName, "Work queue item counts - {0}", _getWorkQueueItemCount());
            }
        }

        /// <summary>
        /// Returns true if work item processing logging is enabled. Users of
        /// this class can enable/disable logging of work item processing by
        /// specifying a delegate via the LogWorkItemProcessing property.
        /// </summary>
        /// <returns></returns>
        private bool _isLogWorkItemProcessingEnabled()
        {
            string methodName = string.Format("[ThreadedWorkQueue<{0}>._logWorkQueueItemCount]", _name);

            try
            {
                if (LogWorkItemProcessing != null)
                {
                    return LogWorkItemProcessing();
                }
            }
            catch (Exception ex)
            {
                _logError(methodName, "Error calling LogWorkItemProcessing delegate - {0}", ex.ToString());
            }

            return true;
        }

        /// <summary>
        /// Logs an error message
        /// </summary>
        /// <param name="source">Who is logging the message.</param>
        /// <param name="format">Message format string.</param>
        /// <param name="formatArgs">Message format string aguments</param>
        private void _logError(object source, string format, params object[] formatArgs)
        {
            Debug.WriteLine(_formatMessage("ERROR", source, format, formatArgs));
        }

        /// <summary>
        /// Logs a warning message
        /// </summary>
        /// <param name="source">Who is logging the message.</param>
        /// <param name="format">Message format string.</param>
        /// <param name="formatArgs">Message format string aguments</param>
        private void _logWarning(object source, string format, params object[] formatArgs)
        {
            Debug.WriteLine(_formatMessage("WARN ", source, format, formatArgs));
        }

        /// <summary>
        /// Logs an informational message
        /// </summary>
        /// <param name="source">Who is logging the message.</param>
        /// <param name="format">Message format string.</param>
        /// <param name="formatArgs">Message format string aguments</param>
        private void _logInformation(object source, string format, params object[] formatArgs)
        {
            //Debug.WriteLine(_formatMessage("INFO ", source, format, formatArgs));
        }

        /// <summary>
        /// Logs a debug message
        /// </summary>
        /// <param name="source">Who is logging the message.</param>
        /// <param name="format">Message format string.</param>
        /// <param name="formatArgs">Message format string aguments</param>
        private void _logDebug(object source, string format, params object[] formatArgs)
        {
            //Debug.WriteLine(_formatMessage("DEBUG", source, format, formatArgs));
        }

        /// <summary>
        /// Logs a fine debug message
        /// </summary>
        /// <param name="source">Who is logging the message.</param>
        /// <param name="format">Message format string.</param>
        /// <param name="formatArgs">Message format string aguments</param>
        private void _logFineDebug(object source, string format, params object[] formatArgs)
        {
            //Debug.WriteLine(_formatMessage("FINE ", source, format, formatArgs));
        }

        /// <summary>
        /// Formats the message for logging.
        /// </summary>
        /// <param name="categoryName">Name of the category in which the message is defined.</param>
        /// <param name="source">Who is logging the message.</param>
        /// <param name="format">Message format string.</param>
        /// <param name="formatArgs">Message format string aguments</param>
        /// <returns></returns>
        private static string _formatMessage(string categoryName, object source, string format, params object[] formatArgs)
        {
            string message = string.Empty;

            try
            {
                message = string.Format(
                                    string.Format(
                                                "{0} - {1} {2}",
                                                categoryName != null ? categoryName : "<null>",
                                                source != null ? source.ToString() : "<null>",
                                                format),
                                    formatArgs);
            }
            catch (Exception ex)
            {
                try
                {
                    // Convert the format arguments to a string for logging
                    StringBuilder args = new StringBuilder();
                    if (formatArgs != null)
                    {
                        for (int i = 0; i < formatArgs.Length; i++)
                        {
                            if (args.Length > 0) args.Append(" ");
                            args.Append(string.Format("{{0}}:'{1}'", i, formatArgs[i] != null ? formatArgs[i].ToString() : "<null>"));
                        }
                    }
                    else
                    {
                        args.Append("<null>");
                    }

                    message = string.Format(
                                "[ThreadedWorkQueue._formatMessage] Error formatting message - category: '{0}' source: '{1}' format: '{2}' formatParams: '{3}': {4}",
                                categoryName != null ? categoryName : "<null>",
                                source != null ? source.ToString() : "<null>",
                                format != null ? format : "<null>",
                                args.ToString(),
                                ex.Message);

                }
                catch (Exception ex1)
                {
                    message = string.Format(
                                "!!! FORMAT ERROR EXCEPTION !!! [ThreadedWorkQueue._formatMessage]" +
                                "Exception handling message format exception - " +
                                "Original message being logged: {0} - " +
                                "Original message format error - {1} " +
                                "Message format exception: {2}",
                                format != null ? format : "<null>",
                                ex.ToString(),
                                ex1.ToString());
                }
            }

            return message;
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Cleanup
        /// </summary>
        public void Dispose()
        {
            ShutdownNow();
            ClearAll();
            _isDisposed = true;
        }

        #region Inner classes

        //---------- _WorkQueueItem -------------------------------------------

        /// <summary>
        /// Class used to represent a prioritized work item on a queue
        /// </summary>
        private class _WorkQueueItem
        {
            internal int Priority { get; private set; }
            internal T WorkItem { get; private set; }
            internal bool Exit { get; set; }

            internal _WorkQueueItem(T workItem, int priority)
            {
                Priority = priority;
                WorkItem = workItem;
                Exit = false;
            }

            public override string ToString()
            {
                string methodName = "[ThreadedWorkQueue._WorkQueueItem.ToString]";

                string str = string.Empty;

                try
                {
                    str = string.Format(
                                    "Priority '{0}' Exit '{1}' WorkItem '({2}) {3}'",
                                    Priority,
                                    Exit,
                                    WorkItem != null ? WorkItem.GetType().ToString() : "<null>",
                                    WorkItem != null ? WorkItem.ToString() : "<null>");

                }
                catch (Exception ex)
                {
                    Debug.WriteLine(_formatMessage("ERROR", methodName, "Exception - {0}", ex.ToString()));
                }

                return str;
            }
        }

        //---------- PriorityDecendingComparer -------------------------------------------

        /// <summary>
        /// This comparer is used to sort work items by priority in a decending order.
        /// The higher priorities will appear at the top and lower priorities
        /// will appear at the bottom.
        /// </summary>
        public class PriorityDecendingComparer : IComparer<int>
        {
            /// <summary>
            /// Sorts priorities in decending order.
            /// </summary>
            /// <param name="x"></param>
            /// <param name="y"></param>
            /// <returns></returns>
            public int Compare(int x, int y)
            {
                if (x == y) return 0;
                if (x < y) return 1;

                return -1;
            }
        }

        #endregion

        #endregion
    }
}
